import asyncio
import os
import random
import signal
import socket
import time
from contextlib import asynccontextmanager, suppress

import pytest
from prometheus_client.openmetrics.exposition import generate_latest

from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.metrics.AmpelProcessCollector import AmpelProcessCollector
from ampel.metrics.prometheus import mmap_dict
from ampel.util.concurrent import _Process, process


def echo(arg):
    time.sleep(random.uniform(0.0005, 0.001))
    return arg


def abort():
    signal.raise_signal(signal.SIGABRT)


def throw():
    raise NotImplementedError


class Unpicklable:
    def __reduce__(self):
        raise NotImplementedError


def return_unpicklable():
    return Unpicklable()


@process
def decorated(arg):
    return arg


@process(timeout=0.1)
def sleepy(dt, stubborn=False):
    if stubborn:
        signal.signal(signal.SIGTERM, lambda stack, n: 0)
    time.sleep(dt)
    return dt


@pytest.mark.asyncio
async def test_launch():
    p = _Process(target=echo, args=(42,))
    result = await p.launch()
    assert result == 42


@pytest.mark.asyncio
async def test_abort():
    p = _Process(target=abort)
    with pytest.raises(RuntimeError):
        await p.launch()


@pytest.mark.asyncio
async def test_raise():
    p = _Process(target=throw)
    with pytest.raises(NotImplementedError):
        await p.launch()


@pytest.mark.asyncio
async def test_too_many_arguments():
    p = _Process(target=throw, args=("borkybork",))
    with pytest.raises(TypeError):
        await p.launch()


@pytest.mark.asyncio
async def test_unpicklable_return():
    p = _Process(target=return_unpicklable)
    with pytest.raises(NotImplementedError):
        await p.launch()


@pytest.mark.asyncio
async def test_decorator():
    assert (await decorated(42)) == 42


@pytest.mark.asyncio
async def test_cancel():
    task = sleepy(5)
    await asyncio.sleep(1)
    task.cancel()
    with pytest.raises(asyncio.exceptions.CancelledError):
        await task


@pytest.mark.asyncio
async def test_kill():
    task = sleepy(10, stubborn=True)
    await asyncio.sleep(1)
    task.cancel()
    t0 = time.time()
    with pytest.raises(asyncio.exceptions.CancelledError):
        await task
    dt = time.time() - t0
    assert dt < 1, "task killed before exiting"


@pytest.mark.asyncio
async def test_multilaunch():
    launch = lambda: asyncio.create_task(_Process(target=echo, args=(42,)).launch())
    count = 0
    num_tasks = 5
    pending = {launch() for _ in range(num_tasks)}
    done = {}
    while True:
        done, pending = await asyncio.wait_for(
            asyncio.wait(pending, return_when="FIRST_COMPLETED"), timeout=num_tasks
        )
        for future in done:
            assert future.result() == 42
            count += 1
        if count >= 20:
            for result in await asyncio.wait_for(
                asyncio.gather(*pending, return_exceptions=True), timeout=num_tasks
            ):
                assert result == 42
            break
        while len(pending) < num_tasks:
            pending.add(launch())


@process
def set_counter(value, process=None):
    AmpelMetricsRegistry.counter(
        "countcount", "cookies", subsystem="test_concurrent"
    ).inc(value)
    AmpelMetricsRegistry.histogram(
        "counthist", "cookies", subsystem="test_concurrent"
    ).observe(1)
    return value


@pytest.fixture
def prometheus_multiproc_dir(monkeypatch, tmpdir):
    """
    Ensure that Prometheus multiprocessing mode is enabled
    """
    from ampel.metrics.AmpelMetricsRegistry import MultiProcessCollector

    monkeypatch.setenv("PROMETHEUS_MULTIPROC_DIR", str(tmpdir))
    r = AmpelMetricsRegistry.registry()
    try:
        c = MultiProcessCollector(r)
        yield tmpdir
    finally:
        r.unregister(c)



@pytest.mark.asyncio
async def test_multiprocess_metrics(prometheus_multiproc_dir):

    sample = lambda: {
        sample.name: sample.value
        for metric in AmpelMetricsRegistry.collect()
        for sample in metric.samples
        if sample
    }
    read_mmap = lambda fname: {
        item[0]: item[1]
        for item in mmap_dict.MmapedDict.read_all_values_from_file(
            prometheus_multiproc_dir / fname
        )
    }
    before = sample()

    value = 1100101
    assert (await set_counter(value, {"name": "0"})) == value

    assert sorted(os.listdir(prometheus_multiproc_dir)) == [
        "counter_archive.db",
        "histogram_archive.db",
    ], "persistent metrics were consolidated"

    after = sample()
    key = "ampel_test_concurrent_countcount_total"
    assert after[key] - before.get(key, 0) == value, "counter was incremented"

    hist_before = read_mmap("histogram_archive.db")
    await set_counter(value, {"name": "0"})
    hist_after = read_mmap("histogram_archive.db")
    assert len(hist_before) == len(
        hist_after
    ), "mmap files have constant size (if labels, including the implicit process label, are constant!)"
    for k in hist_before:
        assert hist_after[k] == hist_before[k] or hist_after[k] == 2 * hist_before[k]


@pytest.mark.asyncio
async def test_multiprocess_metrics_deduplication(prometheus_multiproc_dir):
    """
    Metrics are not duplicated in multiprocess mode
    """

    def get_help_lines():
        expo = generate_latest(AmpelMetricsRegistry)
        return [
            line.split(b" ")[2]
            for line in expo.split(b"\n")
            if line.startswith(b"# HELP")
        ]

    await set_counter(42, {"name": "0"})
    # register metric again in main process
    AmpelMetricsRegistry.counter("countcount", "cookies", subsystem="test_concurrent")

    helps = get_help_lines()
    assert len(set(helps)) == len(helps), "no duplicated HELP lines"


@pytest.mark.asyncio
async def test_implicit_labels(prometheus_multiproc_dir):
    """
    Implicit labels are added if there is a dict argument with a key "name"
    """

    get_sample_value = AmpelMetricsRegistry.registry().get_sample_value
    assert (
        get_sample_value("ampel_test_concurrent_countcount_total", {"name": "hola"})
        is None
    )

    value = 1100101
    assert (await set_counter(value, {"name": "hola"})) == value

    before = {}
    for metric in AmpelMetricsRegistry.registry().collect():
        for sample in metric.samples:
            key = (sample.name, tuple(sample.labels.items()))
            before[key] = sample.value

    assert (
        get_sample_value("ampel_test_concurrent_countcount_total", {"process": "hola"})
        is not None
    )


@asynccontextmanager
async def fence(port):
    """
    Wait for a connection on a TCP port, yield, then reply
    """
    condition = asyncio.Condition()

    async def connected(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        try:
            await reader.read(4)
            async with condition:
                condition.notify()
            async with condition:
                await condition.wait()
            writer.write(b"hola")
            await writer.drain()
            async with condition:
                condition.notify()
        finally:
            writer.close()

    server = await asyncio.start_server(connected, "127.0.0.1", port)
    task = asyncio.create_task(server.serve_forever())

    async with condition:
        await condition.wait()
    try:
        yield
    finally:
        async with condition:
            condition.notify()
        async with condition:
            await condition.wait()

        server.close()
        await server.wait_closed()
        with suppress(asyncio.CancelledError):
            await task


def latch(port):
    """Connect to a port and wait for a reply before exiting"""
    with socket.create_connection(("127.0.0.1", port)) as sock:
        sock.send(b"hola")
        sock.recv(4)


@pytest.mark.asyncio
async def test_process_collector(unused_tcp_port):
    for metric in AmpelProcessCollector().collect():
        assert not metric.samples

    for metric in AmpelProcessCollector(name="self").collect():
        assert len(metric.samples) == 1
        sample = metric.samples[0]
        assert sample.labels == {"process": "self", "replica": "0"}
        assert sample.value > 0

    proc = asyncio.create_task(
        _Process(latch, args=(unused_tcp_port,), name="latch").launch()
    )

    async with fence(unused_tcp_port):
        for metric in AmpelProcessCollector().collect():
            assert len(metric.samples) == 1
            sample = metric.samples[0]
            assert sample.labels == {"process": "latch", "replica": "0"}
            assert sample.value > 0
    await asyncio.wait_for(proc, 3)


@pytest.mark.asyncio
async def test_replica_numbering():
    launch = lambda: _Process(target=echo, args=(42,), name="echo").launch()
    assert not _Process._active
    assert not _Process._expired
    await launch()
    assert not _Process._active
    assert not _Process._expired



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/T3Processor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.02.2018
# Last Modified Date:  03.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Annotated, Any

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.core.EventHandler import EventHandler
from ampel.log import SHOUT, AmpelLogger, LogFlag
from ampel.model.t3.T3IncludeDirective import T3IncludeDirective
from ampel.model.UnitModel import UnitModel
from ampel.struct.T3Store import T3Store
from ampel.types import ChannelId


class T3Processor(AbsEventUnit):

	# Require single channel for now (super classes allow multi-channel)
	channel: None | ChannelId = None

	include: None | T3IncludeDirective

	#: Unit must be a subclass of AbsT3Supplier
	supply: Annotated[UnitModel, AbsT3Supplier]

	#: Unit must be a subclass of AbsT3Stager
	stage: Annotated[UnitModel, AbsT3Stager]


	def post_init(self):
		if self.supply.unit not in self.context.config._config['unit']:  # noqa: SLF001
			raise ValueError(f"Unknown supply unit: {self.supply.unit}")
		if self.stage.unit not in self.context.config._config['unit']:  # noqa: SLF001
			raise ValueError(f"Unknown stager unit: {self.stage.unit}")


	def proceed(self, event_hdlr: EventHandler) -> None:

		event_hdlr.set_tier(3)
		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, event_hdlr.get_run_id(),
			base_flag = LogFlag.T3 | LogFlag.CORE | self.base_log_flag,
			force_refresh = True
		)

		try:

			# Feedback
			logger.log(SHOUT, f'Running {self.process_name}')

			t3s = T3Store()

			if self.include:

				if self.include.docs:
					pass # later

				if (x := self.include.session):
					sdict: dict[str, Any] = {}
					for model in [x] if isinstance(x, UnitModel) else x:
						rd = self.context.loader.new_context_unit(
							model = model,
							context = self.context,
							sub_type = AbsT3Supplier,
							event_hdlr = event_hdlr,
							logger = logger
						).supply(t3s)
						if rd:
							sdict |= rd
					if sdict:
						t3s.add_session_info(sdict)

			supplier = self.context.loader.new_context_unit(
				model = self.supply,
				context = self.context,
				sub_type = AbsT3Supplier,
				logger = logger,
				event_hdlr = event_hdlr
			)

			# Stager unit
			#############

			stager = self.context.loader.new_context_unit(
				model = self.stage,
				context = self.context,
				sub_type = AbsT3Stager,
				logger = logger,
				event_hdlr = event_hdlr,
				channel = (
					config['channel']
					if isinstance((config := self.stage.config), dict) and 'channel' in config
					else self.channel
				)
			)

			logger.info("Running stager", extra={'unit': self.stage.unit})

			if (doc_gen := stager.stage(supplier.supply(t3s), t3s)):
				for t3d in doc_gen:
					if 'meta' not in t3d:
						raise ValueError("Invalid T3Document")
					t3d['meta']['traceid'] = {'t3processor': self._trace_id}
					if event_hdlr.job_sig:
						t3d['meta']['jobid'] = event_hdlr.job_sig
					self.context.db.get_collection('t3').insert_one(t3d)

			"""
			if t3s.resources:
				for v in t3s.resources.values():
					event_hdlr.add_resource(v, overwrite=self.allow_resource_override)

			if t3s.aliases:
				for k, v in t3s.aliases.items():
					event_hdlr.add_alias(k, v, overwrite=self.allow_alias_override)
			"""


		except Exception as e:
			event_hdlr.handle_error(e, logger)

		# Feedback
		logger.log(SHOUT, f'Done running {self.process_name}')
		logger.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/DistConfigBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.10.2019
# Last Modified Date:  18.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import re
from collections.abc import Callable, Sequence
from functools import partial
from importlib import metadata
from typing import Any

import yaml

from ampel.config.builder.ConfigBuilder import ConfigBuilder
from ampel.log import SHOUT, VERBOSE
from ampel.util.distrib import PathLike, PathList, get_dist_names, get_files


class DistConfigBuilder(ConfigBuilder):
	"""
	Subclass of ConfigLoader allowing to automatically detect and load
	various configuration files contained in ampel sub-packages installed with pip
	"""

	def load_distributions(self,
		prefixes: Sequence[str] = ("pyampel-", "ampel-"),
		conf_dirs: Sequence[str] = ("conf",),
		exts: Sequence[str] = ("json", "yaml", "yml"),
		raise_exc: bool = True,
		exclude: None | list[str] = None
	) -> None:
		"""
		:param prefixes: loads all known conf files from all distributions with name starting with prefixes
		:param conf_dirs: loads only conf files from these directories
		:param exts: loads only conf files with these extensions
		:param exclude: exclude distribution by name
		"""
		dist_names = get_dist_names(prefixes)

		if dist_names:
			self.logger.log(SHOUT, f"Detected ampel components: '{dist_names}'")

		if exclude:
			for i in range(len(exclude)):
				exclude[i] = exclude[i].lower()

		for dist_name in dist_names:

			self.logger.break_aggregation()

			if exclude and dist_name.lower() in exclude:
				self.logger.log(VERBOSE, "Excluding distribution '{dist_name}' as requested")
				continue

			s = f"Checking distribution '{dist_name}'"
			self.logger.log(VERBOSE, s)
			self.logger.log(VERBOSE, "="*len(s))

			for conf_dir in conf_dirs:
				for ext in exts:
					self.load_distrib(dist_name, conf_dir, ext, raise_exc=raise_exc)

			if self.verbose:
				self.logger.log(VERBOSE, f"Done checking distribution '{dist_name}'")

		if self.verbose:
			self.logger.log(VERBOSE, "Done loading distributions")


	def load_distrib(self,
		dist_name: str,
		conf_dir: str = "conf",
		ext: str = "json",
		raise_exc: bool = True
	) -> None:
		"""
		Loads all known conf files of the provided distribution (name)
		"""
		try:

			distrib = metadata.distribution(dist_name)
			all_conf_files = get_files(dist_name, conf_dir, re.compile(rf".*\.{ext}$"))

			if all_conf_files and self.verbose:
				self.logger.log(VERBOSE, "Following conf files will be parsed:")
				for el in all_conf_files:
					self.logger.log(VERBOSE, el.as_posix())

			if ampel_conf := self.get_conf_file(all_conf_files, f"ampel.{ext}"):
				self.load_conf_using_func(
					distrib, ampel_conf, self.load_ampel_conf, raise_exc=raise_exc
				)

			# Channel, mongo (and template) can be defined by multiple files
			# in a directory named after the corresponding config section name
			for key in ("channel", "mongo", "process"):
				if specialized_conf_files := self.get_conf_files(all_conf_files, f"*/{key}/*"):
					for f in specialized_conf_files:
						self.load_conf_using_func(distrib, f, self.first_pass_config[key].add)

			# ("channel", "mongo", "resource")
			for key in self.first_pass_config.conf_keys:
				if key == "alias":
					continue
				if section_conf_file := self.get_conf_file(all_conf_files, f"{key}.{ext}"):
					self.load_conf_using_func(distrib, section_conf_file, self.first_pass_config[key].add)

			# ("controller", "processor", "unit", "alias", "process")
			for unit_type in ("alias", "process"):
				if tier_conf_file := self.get_conf_file(all_conf_files, f"{unit_type}.{ext}"):
					self.load_conf_using_func(
						distrib, tier_conf_file, partial(self.register_tier_conf, unit_type)
					)

			# Try to load templates from folder template (defined by 'Ampel-ZTF' for ex.)
			if template_conf_files := self.get_conf_files(all_conf_files, "*/template/*"):
				for f in template_conf_files:
					self.load_conf_using_func(distrib, f, self.register_templates)

			# Try to load templates from template.conf
			if template_conf := self.get_conf_file(all_conf_files, "template.{ext}"):
				self.load_conf_using_func(distrib, template_conf, self.register_templates)

			if all_conf_files:
				self.logger.info(f"Not all conf files were loaded from distribution '{distrib.name}'")
				self.logger.info(f"Unprocessed conf files: {all_conf_files}")

		except Exception as e:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"Error occured while loading configuration files from the distribution '{dist_name}'",
				exc_info=e
			)


	def register_tier_conf(self,
		root_key: str,
		d: dict[str, Any],
		dist_name: str,
		version: str,
		file_rel_path: str,
	):
		for k in ("t0", "t1", "t2", "t3", "ops"):
			if k in d:
				self.first_pass_config[root_key][k].add(
					d[k],
					dist_name = dist_name,
					version = version,
					register_file = file_rel_path,
				)


	def load_conf_using_func(self,
		distrib: metadata.Distribution,
		file_rel_path: "PathLike",
		func: Callable[[dict[str, Any], str, str, str], None],
		raise_exc: bool = True,
		**kwargs
	) -> None:

		try:

			if self.verbose:
				self.logger.log(VERBOSE,
					f"Loading {file_rel_path} from distribution '{distrib.name}'"
				)

			# NB: read both YAML and JSON (which is a subset of YAML)
			content = yaml.safe_load(file_rel_path.read_text())

			func(
				content,
				distrib.name,
				distrib.version,
				file_rel_path.as_posix(),
				**kwargs
			)

		except FileNotFoundError:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"File '{file_rel_path}' not found in distribution '{distrib.name}'"
			)

		except Exception as e:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"Error occured loading '{file_rel_path}' from distribution '{distrib.name}'",
				exc_info=e
			)


	@staticmethod
	def get_conf_file(files: "PathList", key: str) -> "None | PathLike":
		"""
		Extract the first entry who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched element)
		"""
		for el in files:
			if el.match(key):
				files.remove(el)
				return el
		return None


	@staticmethod
	def get_conf_files(files: "PathList", key: str) -> "PathList":
		"""
		Extract all entries who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched elements)
		"""
		ret = [el for el in files if el.match(key)]
		for el in ret:
			files.remove(el)
		return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t4/T4RunTimeContextUpdater.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.04.2023
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator, Sequence
from time import time
from typing import Annotated, Any

from ampel.abstract.AbsT4ControlUnit import AbsT4ControlUnit
from ampel.abstract.AbsT4Unit import AbsT4Unit
from ampel.content.T4Document import T4Document
from ampel.core.DocBuilder import DocBuilder
from ampel.model.UnitModel import UnitModel


class T4RunTimeContextUpdater(AbsT4ControlUnit, DocBuilder):

	execute: Annotated[Sequence[UnitModel], AbsT4Unit]
	allow_alias_override: bool = False

	def __init__(self, **kwargs) -> None:
		if isinstance(kwargs.get('execute'), dict):
			kwargs['execute'] = [kwargs['execute']]
		super().__init__(**kwargs)
			

	def do(self) -> Generator[T4Document, None, None]:

		aliases: dict[str, Any] = {}
		ts = time()
		for um in self.execute:
			t4_unit = self.context.loader.new_safe_logical_unit(
				um=um, unit_type=AbsT4Unit, logger=self.logger,
				_chan=self.channel # type: ignore[arg-type]
			)
			if ret := t4_unit.do():
				if not (isinstance(ret, dict) and all(isinstance(k, str) for k in ret)):
					raise ValueError(f'Invalid {um.unit} return value, dict[str, Any] expected')
				# Ensure alias keys start with %%
				ret = {k if k.startswith('%%') else f'%%{k}': v for k, v in ret.items()}
				for k in ret:
					if not self.allow_alias_override and k in aliases:
						raise ValueError(
							f'Run time alias {k} was already registered, '
							f'set allow_alias_override=True to ignore'
						)
				aliases |= ret

		for k, v in aliases.items():
			self.context.add_run_time_alias(k, v)

		yield self.craft_doc(self.event_hdlr, self, aliases, ts, doc_type=T4Document)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoT4RunTimeAliasGenerator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.12.2022
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from random import randint

from ampel.abstract.AbsT4Unit import AbsT4Unit
from ampel.struct.UnitResult import UnitResult
from ampel.types import UBson


class DemoT4RunTimeAliasGenerator(AbsT4Unit):
	""" To be run with T4RunTimeContextUpdater """

	debug: bool = False
	alias_name: str = "%%steven"

	def post_init(self) -> None:
		self.logger.info("post_init was called")

	def do(self) -> UBson | UnitResult:
		self.logger.info(f"Running {self.__class__.__name__}")
		return {self.alias_name: randint(10, 30)}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/time/TimeLastRunModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  31.07.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from datetime import datetime, timedelta, timezone
from typing import Literal

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.core.AmpelDB import AmpelDB
from ampel.mongo.query.var.events import get_last_run


class TimeLastRunModel(AmpelBaseModel):

	match_type: Literal['time_last_run']
	process_name: str
	fallback: None | dict = {'days': -1}
	require_success: bool = True


	def get_timestamp(self, **kwargs) -> None | float:
		now = kwargs.get('now') or datetime.now(tz=timezone.utc)
		if ts := self._query_events_col(kwargs['db'], now, self):
			return ts
		if self.fallback:
			return (now + timedelta(**self.fallback)).timestamp()
		return None


	@staticmethod
	def _query_events_col(ampel_db: AmpelDB, now: datetime, model: 'TimeLastRunModel') -> None | float:

		col = ampel_db.get_collection('event')

		# First query the last 10 days
		res = get_last_run(
			col, model.process_name, model.require_success,
			gte_time=(now - timedelta(days=10)).timestamp(),
			timestamp = True,
		)

		# If nothing is found, try querying the entire collection (days_back=None)
		if res is None:
			res = get_last_run(col, model.process_name, model.require_success, gte_time=None, timestamp=True)
			if not res:
				return None

		return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT3Supplier.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.07.2021
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler
from ampel.log.AmpelLogger import AmpelLogger
from ampel.struct.T3Store import T3Store
from ampel.types import T, Traceless


class AbsT3Supplier(Generic[T], AmpelABC, ContextUnit, abstract=True):
	"""
	Abstract class for T3 suppliers
	"""

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]

	@abstractmethod
	def supply(self, t3s: T3Store) -> T:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/model/AmpelColModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.10.2019
# Last Modified Date:  11.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.model.IndexModel import IndexModel
from ampel.mongo.model.ShortIndexModel import ShortIndexModel


class AmpelColModel(AmpelBaseModel):
	name: str
	indexes: None | Sequence[ShortIndexModel | IndexModel] = None
	args: dict = {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/alter/ResolveRunTimeAliases.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.04.2023
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.abstract.AbsConfigUpdater import AbsConfigUpdater
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.recursion import walk_and_process_dict


class ResolveRunTimeAliases(AbsConfigUpdater):

	def alter(self, context: AmpelContext, content: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:

		# Run-time aliases generated at T4 for T0/T1 processes
		if context.run_time_aliases:
			walk_and_process_dict(
				arg = content,
				callback = self._gather_run_time_aliases_callback,
				run_time_aliases = context.run_time_aliases,
				logger = logger
			)

		return content

	def _gather_run_time_aliases_callback(self, path, current_key, current_d, **kwargs) -> None:
		""" Used by walk_and_process_dict(...) from morph(...) """
		# print(f"# path: {path}\n# d: {d}\n")
		for k, v in current_d.items():
			if isinstance(v, str) and v[0] == '%' == v[1]:
				for rt_key in kwargs['run_time_aliases']:
					if v == rt_key:
						if kwargs['logger'].verbose:
							kwargs['logger'].info(
								f"Setting value for run time alias {rt_key} with path {path}.{k}"
							)
						current_d[k] = kwargs['run_time_aliases'][rt_key]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/AbsCoreCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.03.2021
# Last Modified Date:  11.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
import os
import re
from collections.abc import Iterator, Sequence
from typing import Any, Literal, TypeVar

from ampel.abstract.AbsCLIOperation import AbsCLIOperation
from ampel.cli.config import get_user_data_config_path
from ampel.cli.utils import _maybe_int, get_db, get_vault
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.AmpelContext import AmpelContext
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.freeze import recursive_freeze
from ampel.util.mappings import set_by_path
from ampel.util.pretty import out_stack

custom_conf_patter = re.compile(r"^--[\w-]*(?:\.[\w-]+)*.*$")

T = TypeVar("T", bound = AmpelContext)

class AbsCoreCommand(AbsCLIOperation, abstract=True):

	def __init__(self):
		self.parsers = {}
		try:
			import sys

			import IPython
			sys.breakpointhook = IPython.embed
		except Exception:
			pass


	def load_config(self,
		config_path: None | str,
		unknown_args: Sequence[str],
		logger: None | AmpelLogger = None,
		freeze: bool = True,
		env_var_prefix: None | str = "AMPEL_CONFIG_",
	) -> AmpelConfig:

		if not config_path:
			std_conf = get_user_data_config_path()
			if os.path.exists(std_conf):
				ampel_conf = AmpelConfig.load(std_conf, freeze=False)
			else:
				with out_stack():
					raise ValueError("No default ampel config found -> argument -config required\n")
		else:
			ampel_conf = AmpelConfig.load(config_path, freeze=False)

		if logger is None:
			logger = AmpelLogger.get_logger()

		if env_var_prefix is not None:
			for var, value in os.environ.items():
				if var.startswith(env_var_prefix):
					k = var[len(env_var_prefix):]
					try:
						v = json.loads(value)
					except json.JSONDecodeError:
						v = value
					logger.info(f"Setting config parameter '{k}' from environment")
					set_by_path(ampel_conf._config, k, v)  # noqa: SLF001

		for k, v in self.get_custom_args(unknown_args):

			if ampel_conf.get(".".join(k.split(".")[:-1])) is None:
				with out_stack():
					raise ValueError(f"Unknown config parameter '{k}'\n")

			logger.info(f"Setting config parameter '{k}' value to: {v}")
			set_by_path(ampel_conf._config, k, v)  # noqa: SLF001

		if freeze:
			ampel_conf._config = recursive_freeze(ampel_conf._config)  # noqa: SLF001

		return ampel_conf


	def get_custom_args(self, customizations: Sequence[str]) -> Iterator[tuple[str, Any]]:
		"""
		:raises: ValueError
		"""

		it = iter(customizations)
		for el in it:
			if custom_conf_patter.match(el):
				if "=" in el:
					s = el.split("=")
					k = s[0][2:]
					v = _maybe_int(s[1])
				else:
					k = el[2:]
					try:
						v = _maybe_int(next(it))
					except StopIteration:
						v = None
				yield k, v
			else:
				with out_stack():
					raise ValueError(f"Unknown argument: {el}\n")


	def get_context(self,
		args: dict[str, Any],
		unknown_args: Sequence[str],
		logger: None | AmpelLogger = None,
		freeze_config: bool = True,
		ContextClass: type[T] = AmpelContext, # type: ignore[assignment]
		require_existing_db: bool | str = True,
		one_db: bool | Literal['auto'] = False,
		**kwargs
	) -> T:
		"""
		:require_existing_db: str typed values specify required database prefix
		"""

		if logger is None:
			logger = AmpelLogger.get_logger()

		config = self.load_config(
			args['config'], unknown_args, logger, freeze = freeze_config
		)

		vault = get_vault(args)
		db = get_db(config, vault, require_existing_db, one_db)
		return ContextClass(
			config = config,
			db = db,
			loader = UnitLoader(config, db=db, vault=vault),
			**kwargs
		)


	def convert_logical_args(self, name: str, args: dict[str, Any]) -> None:

		for k in (f"with_{name}", f"with_{name}s_and", f"with_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['with'] = args[k]
				break

		for k in (f"without_{name}", f"without_{name}s_and", f"without_{name}s_or"):
			if args.get(k):
				if name not in args:
					args[name] = {}
				args[name]['without'] = args[k]
				break



import os
from time import perf_counter_ns
from typing import Any, ClassVar

from prometheus_client import (
    CollectorRegistry,
    Counter,
    Gauge,
    Histogram,
    Summary,
    push_to_gateway,
)
from prometheus_client.context_managers import Timer
from prometheus_client.metrics import MetricWrapperBase
from prometheus_client.multiprocess import MultiProcessCollector


def reset_metric(metric: MetricWrapperBase) -> None:
    if metric._is_parent():  # noqa: SLF001
        for child in metric._metrics.values():  # noqa: SLF001
            reset_metric(child)
    elif isinstance(metric, Counter | Gauge):
        metric._value.set(0)  # noqa: SLF001
    elif isinstance(metric, Summary):
        metric._sum.set(0)  # noqa: SLF001
        metric._count.set(0)  # noqa: SLF001
    elif isinstance(metric, Histogram):
        metric._sum.set(0)  # noqa: SLF001
        for bucket in metric._buckets:  # noqa: SLF001
            bucket.set(0)
    else:
        raise TypeError(f"don't know how to reset metric of type {type(metric)}")


def reset_registry(registry: CollectorRegistry) -> None:
    for metric in registry._names_to_collectors.values():  # noqa: SLF001
        if isinstance(metric, MetricWrapperBase):
            reset_metric(metric)


class UnitTimer(Timer):
    """
    Our own version of timer that supports a custom time function (mainly for units)
    """

    def __init__(self, metric, callback_name, unit: int = 1000):
        super().__init__(metric, callback_name)
        self._unit = unit

    def timer(self) -> int:
        return perf_counter_ns() // self._unit

    def _new_timer(self):
        return self.__class__(self._metric, self._callback_name, self._unit)

    def __enter__(self):
        self._start = self.timer()
        return self

    def __exit__(self, typ, value, traceback):
        # Time can go backwards.
        duration = max(self.timer() - self._start, 0)
        callback = getattr(self._metric, self._callback_name)
        callback(duration)


class UnitTimable(MetricWrapperBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ns_per_time_unit = None
        if self._unit in ("seconds", "s"):
            self._ns_per_time_unit = 1_000_000_000
        elif self._unit in ("milliseconds", "millis", "ms"):
            self._ns_per_time_unit = 1_000_000
        elif self._unit in ("microseconds", "micros", "µs"):
            self._ns_per_time_unit = 1_000
        elif self._unit in ("nanoseconds", "nanos", "ns"):
            self._ns_per_time_unit = 1


# A counter that knows how to time
class TimingCounter(Counter, UnitTimable):
    def time(self) -> UnitTimer:
        if self._ns_per_time_unit is None:
            raise ValueError(f"Can't time in units of {self._unit}")
        return UnitTimer(self, "inc", self._ns_per_time_unit)


# A summary that knows how to time in the specific unit
class TimingSummary(Summary, UnitTimable):
    def time(self) -> UnitTimer:
        if self._ns_per_time_unit is None:
            raise ValueError(f"Can't time in units of {self._unit}")
        return UnitTimer(self, "observe", self._ns_per_time_unit)


class AmpelMetricsRegistry(CollectorRegistry):

    _registry: ClassVar[None | CollectorRegistry] = None
    _standalone_collectors: ClassVar[list[Any]] = []

    @classmethod
    def registry(cls) -> CollectorRegistry:
        if cls._registry is None:
            cls._registry = CollectorRegistry()
        return cls._registry

    @classmethod
    def push(
        cls, gateway: str, job: str, timeout: float | None = 30, reset: bool = False
    ):
        registry = cls.registry()
        push_to_gateway(gateway, job, registry, timeout=timeout)
        if reset:
            reset_registry(registry)

    @classmethod
    def collect(cls):
        if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
            yield from MultiProcessCollector(None).collect()
        else:
            yield from cls.registry().collect()
        for collector in cls._standalone_collectors:
            yield from collector.collect()

    @classmethod
    def register_collector(cls, collector):
        cls._standalone_collectors.append(collector)

    @classmethod
    def deregister_collector(cls, collector):
        cls._standalone_collectors.remove(collector)

    @classmethod
    def counter(
        cls, name, documentation, unit="", labelnames=(), subsystem=""
    ) -> TimingCounter:
        return TimingCounter(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
        )

    @classmethod
    def summary(
        cls, name, documentation, unit="", labelnames=(), subsystem=""
    ) -> Summary:
        return TimingSummary(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
        )

    @classmethod
    def gauge(
        cls,
        name,
        documentation,
        unit="",
        labelnames=(),
        subsystem="",
        multiprocess_mode="livesum",
    ) -> Gauge:
        return Gauge(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
            multiprocess_mode=multiprocess_mode,
        )

    @classmethod
    def histogram(
        cls,
        name,
        documentation,
        unit="",
        labelnames=(),
        subsystem="",
        buckets=Histogram.DEFAULT_BUCKETS,
    ) -> Histogram:
        return Histogram(
            name,
            documentation,
            unit=unit,
            labelnames=labelnames,
            namespace="ampel",
            subsystem=subsystem,
            registry=cls.registry(),
            buckets=buckets,
        )



from contextlib import contextmanager

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsIngester import AbsIngester
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.content.DataPoint import DataPoint
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


class MongoIngester(AbsIngester):
    updates_buffer_size: int = 500

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        updates_buffer = DBUpdatesBuffer(
            self.context.db,
            self.run_id,
            self.logger,
            error_callback=self.error_callback,
            acknowledge_callback=self.acknowledge_callback,
            catch_signals=False,  # we do it ourself
            max_size=self.updates_buffer_size,
            raise_exc=self.raise_exc,
        )

        # Create ingesters
        dbconf = self.context.config.get("mongo.ingest", dict, raise_exc=True)

        def get_ingester_model(key: str) -> UnitModel:
            model = dbconf[key]
            if isinstance(model, str):
                return UnitModel(unit=model)
            return UnitModel(**model)

        self._t0 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t0"),
            sub_type=AbsDocIngester[DataPoint],
            updates_buffer=updates_buffer,
        )

        self._t1 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t1"),
            sub_type=AbsDocIngester[T1Document],
            updates_buffer=updates_buffer,
        )

        self._t2 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t2"),
            sub_type=AbsDocIngester[T2Document],
            updates_buffer=updates_buffer,
        )

        self._stock = AuxUnitRegister.new_unit(
            model=get_ingester_model("stock"),
            sub_type=AbsDocIngester[StockDocument],
            updates_buffer=updates_buffer,
        )

        updates_buffer.start()

        self.updates_buffer = updates_buffer

    def __del__(self) -> None:
        self.flush()

    @contextmanager
    def group(self):
        with self.updates_buffer.group_updates():
            yield

    def acknowledge_on_delivery(self, message):
        self.updates_buffer.acknowledge_on_push(message)

    def flush(self):
        self.updates_buffer.stop()
        self.updates_buffer.push_updates(force=True)

    @property
    def stock(self) -> AbsDocIngester[StockDocument]:
        return self._stock

    @property
    def t0(self) -> AbsDocIngester[DataPoint]:
        return self._t0

    @property
    def t1(self) -> AbsDocIngester[T1Document]:
        return self._t1

    @property
    def t2(self) -> AbsDocIngester[T2Document]:
        return self._t2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/view/MongoChannelView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.03.2021
# Last Modified Date:  26.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.decorator import abstractmethod


# mypy: disable-error-code = empty-body
class AbsMongoView(AmpelABC, AmpelBaseModel, abstract=True):

	@abstractmethod
	def stock(self) -> list[dict[str, Any]]:
		...

	@abstractmethod
	def t0(self) -> list[dict[str, Any]]:
		...

	@abstractmethod
	def t1(self) -> list[dict[str, Any]]:
		...

	@abstractmethod
	def t2(self) -> list[dict[str, Any]]:
		...

	@abstractmethod
	def t3(self) -> list[dict[str, Any]]:
		...

	@abstractmethod
	def get_meta_cases(self, arg: str) -> list[dict[str, Any]]:
		...

	def conform_meta(self) -> dict[str, Any]:
		""" To be overriden if need be """

		return {
			'$filter': {
				'input': {
					"$map": {
						"input": "$meta",
						'as': "metad",
						"in": {
							"$mergeObjects": [
								"$$metad",
								{
									"activity": {
										'$filter': {
											'input': {
												"$map": {
													"input": "$$metad.activity",
													'as': "activityd",
													"in": {
														"$switch": {
															"branches": [
																{
																	# Handles meta activities without channel
																	'case': {'$lte': ["$$activityd.channel", None]},
																	'then': "$$activityd"
																},
																*self.get_meta_cases("$$activityd")
															],
															"default": False
														}
													}
												}
											},
											"as": "ad",
											"cond": "$$ad"
										}
									}
								}
							]
						}
					}
				},
				"as": "md",
				"cond": "$$md.activity"
			}
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/LogsDumper.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  17.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
import sys
from collections.abc import Sequence
from typing import IO

from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.base.AmpelFlexModel import AmpelFlexModel
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.LogFlag import LogFlag
from ampel.util.pretty import prettyjson
from ampel.util.serialize import walk_and_encode


class LogsDumper(AmpelFlexModel):
	""" later """

	out: None | str = None
	to_json: None | bool = False
	to_pretty_json: None | bool = False
	date_format: None | str = None
	id_mapper: None | AbsIdMapper = None
	datetime_key: str = '_id'
	resolve_flag: bool = True
	main_separator: str = ' '
	extra_separator: str = ' '


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self._flag_strings: dict = {}


	def process(self, log_entries: Sequence[dict]):

		fd = open(self.out, "w") if self.out else sys.stdout # noqa: SIM115

		if self.to_json or self.to_pretty_json:
			self.write_json(fd, log_entries)
		else:
			self.write_txt(fd, log_entries)

		fd.flush()

		if self.out:
			fd.close()


	def write_json(self, fd: IO, log_entries: Sequence[dict]) -> None:

		func = json.dumps if self.to_json else prettyjson

		fd.write("[\n")
		buf = ""
		overwrite_pkey = self.datetime_key != '_id'

		for el in log_entries:
	
			# ObjectId is not json serializable
			# (and not interesting when another field contains a timestamp)
			if overwrite_pkey:
				el['_id'] = el[self.datetime_key]
				del el[self.datetime_key]

			if self.date_format:
				el['_id'] = el['_id'].strftime(self.date_format)

			if 's' in el and self.id_mapper:
				el['s'] = self.id_mapper.to_ext_id(el['s'])

			if isinstance(el['f'], LogFlag):
				if el['f'] not in self._flag_strings:
					self._flag_strings[el['f']] = str(el['f']).replace("LogFlag.", "")
				el['f'] = self._flag_strings[el['f']]

			fd.write(buf)
			walk_and_encode(el)
			buf = func(el) + ",\n"

		fd.write(buf[:-2] + "\n]\n")


	def write_txt(self, fd: IO, log_entries: Sequence[dict]) -> None:

		for el in log_entries:

			out = el[self.datetime_key].strftime(self.date_format) if self.date_format else el[self.datetime_key]

			if isinstance(el['f'], LogFlag):
				if el['f'] not in self._flag_strings:
					self._flag_strings[el['f']] = self.main_separator + str(el['f']).replace("LogFlag.", "")
				out += self._flag_strings[el['f']]
			else:
				out += self.main_separator + str(el['f'])

			suffix = [f"run={el['r']}"]

			if 's' in el:
				if self.id_mapper:
					suffix.append(f"stock={self.id_mapper.to_ext_id(el['s'])}")
				else:
					suffix.append(f"stock={el['s']}")

			if 'c' in el:
				suffix.append(f"channel={el['c']}")

			if 'a' in el:
				suffix.append(f"alert={el['a']}")

			if (e := el.get('e')):
				suffix = [f'{k}={e[k]}' for k in e]

			if 'n' in el:
				suffix.append("new")

			if suffix:
				out += self.main_separator + f'[{self.extra_separator.join(suffix)}]'

			if el.get('m'):
				out += self.main_separator + str(el['m'])

			fd.write(out + "\n")


	def log_entries(self, log_entries: Sequence[dict], logger: AmpelLogger) -> None:
		"""
		Unsure when this could ever be required but it's there just in case
		"""

		for el in log_entries:

			record = LightLogRecord(name=0, levelno=el['f'], msg=el.get('m'))
			record.extra = el.get('e') or {}
			record.extra['run'] = el['r']

			if 'c' in el:
				record.channel = el['c']

			if 's' in el:
				record.stock = el['s']

			if 'a' in el:
				record.extra['alert'] = el['a']

			if 'n' in el:
				record.extra['new'] = True

			logger.handle(record)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/t3/T3IncludeDirective.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.12.2021
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.UnitModel import UnitModel
from ampel.types import OneOrMany


class T3IncludeDirective(AmpelBaseModel):
	"""
	:param session: models for AbsT3Supplier[dict] instances which populates the 'session' field of T3Store
	Examples of session information are:
	- Date and time the current process was last run
	- Number of alerts processed since then
	"""

	#: Provides Iterable[T3Document]
	docs: None | UnitModel = None

	#: Provides session information. Unit(s) must be a subclass of AbsT3Supplier
	session: None | OneOrMany[UnitModel] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/t1.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  20.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections
from collections.abc import Sequence
from typing import Any

from bson.int64 import Int64

from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.mongo.query.general import type_stock_id
from ampel.mongo.schema import apply_schema
from ampel.types import ChannelId, StockId, StrictIterable
from ampel.util.collections import check_seq_inner_type

# ruff: noqa: RUF002

"""
Lastest by mean of compound 'body' length.
Validity requirement: compound size should always increase with time.
Especially: if a compound member is discarded, it should not be deleted
"""

def latest_fast_query(
	stock: StockId | StrictIterable[StockId],
	channel: None | ChannelId | dict | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None
) -> list[dict]:
	"""
	| **Must be used on transients whose compounds were solely created by T0** (i.e with no T3 compounds)
	| Should perform faster than general_query.

	:param stock: transient id(s), query can/should be performed on multiple ids at once.

	:param channel: a single channel or a dict schema.
	None (no criterium) means all channel are considered.

	:returns: a dict instance to be used with the mongoDB **aggregation** framework.

	Example:\n
	.. sourcecode:: python\n
		In []: list(col.aggregate(fast_query(to_ampel_id(['ZTF18aaayyuq', 'ZTF17aaagvng', 'ZTF18aaabikt']))))
		Out[]: [
			{'_id': b'T6TG\x96\x80\x1d\x86\x9f\x11\xf2G\xe7\xf4\xe0\xc3', 'stock': 'ZTF18aaayyuq'},
			{'_id': b'\xaaL|\x94?\xa4\xa1D\xbe\x0c[D\x9b\xc6\xe6o', 'stock': 'ZTF18aaabikt'},
			{'_id': b'\xaaL|\x14?\xb4\xc3D\xd2\x2a?L\x9a\xa6\xa1o', 'stock': 'ZTF17aaagvng'}
		]
	"""

	# Robustness
	if isinstance(stock, Sequence):
		if not check_seq_inner_type(stock, type_stock_id):
			raise ValueError("Elements in stock must be of type str or int or Int64 (bson)")
	elif not isinstance(stock, type_stock_id):
		raise ValueError("stock must be of type str or int or Int64 (or sequence of these types)")

	query = {
		'stock': stock if isinstance(stock, type_stock_id) \
			else {'$in': stock if isinstance(stock, list) else list(stock)}
	}

	if channel is not None:
		apply_schema(query, 'channel', channel)

	return [
		{
			'$match': query
		},
		{
			'$project': {
				'stock': 1,
				'len': 1
			}
		},
		{
			'$sort': {
				'stock': 1,
				'len': -1
			}
		},
		{
			'$group': {
				'_id': '$stock',
				'body': {
					'$first': '$$ROOT'
				}
			}
		},
		{
			'$replaceRoot': {
				'newRoot': '$body'
			}
		},
		{
			'$project': {
				'len': 0
			}
		}
	]


def latest_general_query(
	single_stock: StockId,
	project: None | dict[str, Any] = None,
	channel: None | ChannelId | dict | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None
) -> list[dict[str, Any]]:
	"""
	| Should work with any ampel transients.
	| A detailed explanation of each step of the aggregation \
	is documented in the python notebook "T3 get_lastest_compound"

	:param single_stock: transient id (int, Int64 or string). \
	Query **CANNOT** be performed on multiple ids at once.

	:param dict project: optional projection stage at the end of the aggregation

	:param channel: single channel or a dict schema \
	(see :obj:`QueryMatchSchema <ampel.query.QueryMatchSchema>` for details. \
	None (no criterium) means all channel are considered.

	:returns: A dict instance intended to be used with the mongoDB **aggregation** framework.

	IMPORTANT NOTE: the following two examples show the output of the aggregation framework \
	from mongodb (i.e the output after having performed a DB query **using the output \
	of this fuunction as parameter**), they do not show the output of this function.

	**MONGODB** Output example 1:\n
	.. sourcecode:: python\n
		In []: list(col.aggregate(general_query('ZTF18aaayyuq')))
		Out[]: [
			{
				'_id': b'T6TG\x96\x80\x1d\x86\x9f\x11\xf2G\xe7\xf4\xe0\xc3',
				'added': 1520796310.496276,
				'channel': ['HU_SN1'],
				'lastppdt': 2458158.7708565,
				'len': 12,
				'body': [
					{'id': 375300016315010040},
					{'id': 375320176315010034},
					{'id': 375337116315010046},
					{'id': 375356366315010056},
					{'id': 377293446315010009},
					{'id': 377313156315010027},
					{'id': 377334096315010020},
					{'id': 377376126315010004},
					{'id': 377416496315010000},
					{'id': 378293006315010001},
					{'id': 378334946315010000},
					{'id': 404270856315015007}
				],
				'tier': 0,
				'stock': 'ZTF18aaayyuq'
			}
		]

	**MONGODB** Output example 2:\n
	.. sourcecode:: python\n
		In []: list(col.aggregate(general_query('ZTF18aaayyuq', project={'$project': {'stock':1}})))
		Out[]: [{'_id': '5de2480f28bfca0bd3baae890cb2d2ae', 'stock': 'ZTF18aaayyuq'}]
	"""

	# Robustness
	if isinstance(single_stock, collections.abc.Sequence) or \
		not isinstance(single_stock, int | Int64 | str):
		raise ValueError(
			"Type of single_stock must be a string or an int (multi single_stock queries not supported)"
		)

	query = {'stock': single_stock}

	if channel is not None:
		apply_schema(query, 'channel', channel)

	ret: list[dict[str, Any]] = [
		{
			'$match': query
		},
		{
			'$group': {
				'_id': '$tier',
				'latestAdded': {'$max': '$added'},
				'comp': {'$push': '$$ROOT'}
			}
		},
		{
			'$sort': {'latestAdded': -1}
		},
		{
			'$limit': 1
		},
		{
			'$unwind': '$comp'
		},
		{
			'$project': {
				'_id': 0,
				'comp': 1,
				'sortValueUsed': {
					'$cond': {
						'if': {'$eq': ['$comp.tier', 0]},
						'then': '$comp.len',
						'else': '$comp.added'
					}
				}
			}
		},
		{
			'$sort': {'sortValueUsed': -1}
		},
		{
			'$limit': 1
		},
		{
			'$replaceRoot': {'newRoot': '$comp'}
		}
	]

	if project is not None:
		ret.append(project)

	return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT3Loader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.12.2019
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Iterable, Iterator, Sequence

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.core.DataLoader import DataLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.types import ChannelId, StockId, StrictIterable, Traceless


# Data loaders need access to the ampel db (and hence inherits ContextUnit)
class AbsT3Loader(AmpelABC, ContextUnit, abstract=True):
	"""
	Base class for loading documents associated with a set of stocks.
	"""

	logger: Traceless[AmpelLogger]

	#: Specification of documents to load. If these are supplied as strings,
	#: they will be resolved by retrieving the corresponding alias from the
	#: Ampel config.
	directives: Sequence[LoaderDirective]

	#: Channels to load documents for
	channel: None | ChannelId | AnyOf[ChannelId] | AllOf[ChannelId] | OneOf[ChannelId]


	def __init__(self, **kwargs) -> None:

		# Note: 'directives' in kwargs can contain strings which will be
		# resolved by retrieving the associated alias from the ampel config
		directives: list[dict] = []

		# Resolve directive aliases
		for el in kwargs.get('directives', []):
			if isinstance(el, str):
				d = kwargs['context'].config.get(f"alias.t3.%{el}", dict)
				if d: # mypy does not yet support type inference using the walrus operator
					directives.append(d)
				else:
					raise ValueError(f"LoaderDirective alias '{el}' not found in ampel config")
			else:
				directives.append(el)

		kwargs['directives'] = tuple(directives)

		# No need to save context as instance variable
		super().__init__(**kwargs)

		self.data_loader = DataLoader(self.context)


	@abstractmethod
	def load(self,
		stock_ids: StockId | Iterator[StockId] | StrictIterable[StockId]
	) -> Iterable[AmpelBuffer]:
		""" Load documents (collection Ampel_data) for the selected stocks """
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/logging/handlers/RecordBufferingHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.09.2018
# Last Modified Date:  09.05.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from logging import WARNING, LogRecord

from ampel.log.LightLogRecord import LightLogRecord


class RecordBufferingHandler:
	"""
	MemoryHandler-like class that can grown infinitely.
	The standard memory handler provided by logging (BufferingHandler) makes use of a value called 'capacity',
	which once reached, triggers the flush() method when new log records are emitted.
	Since we trust ourselves to do things right (to never let the buffer grow indefinitely),
	we renounce to use such security measure.

	Known subclasses: DefaultRecordBufferingHandler, ChanRecordBufHandler, EnclosedChanRecordBufHandler
	"""

	__slots__ = 'buffer', 'level', 'has_error', 'warn_lvl'

	def __init__(self, level: int) -> None:
		self.buffer: list[LogRecord | LightLogRecord] = []
		self.level = level
		self.has_error = False
		self.warn_lvl = WARNING


	def flush(self) -> None:
		""" Flush just means erase existing log records """
		self.buffer = []
		self.has_error = False


	def handle(self, record: LogRecord | LightLogRecord) -> None:
		if record.levelno >= self.level:
			self.buffer.append(record)
			if record.levelno > self.warn_lvl:
				self.has_error = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/DocBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.04.2023
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from datetime import datetime, timezone
from typing import Literal, TypeVar

from ampel.abstract.AbsUnitResultAdapter import AbsUnitResultAdapter
from ampel.base.AmpelUnit import AmpelUnit
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T3Document import T3Document
from ampel.content.T4Document import T4Document
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.model.UnitModel import UnitModel
from ampel.struct.UnitResult import UnitResult
from ampel.types import ChannelId, OneOrMany, Tag, UBson, ubson
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.tag import merge_tags

T = TypeVar("T", T3Document, T4Document)


class DocBuilder(ContextUnit):

	channel: None | OneOrMany[ChannelId] = None

	#: If true, value of T3/T4Document.config will be the config dict rather than its hash
	resolve_config: bool = False

	#: Tag(s) to be added to t3/t4 documents if applicable (if unit returns something)
	tag: None | OneOrMany[Tag] = None

	#: If set, value of T3/T4Document._id will be built using the 'elements' listed below.
	human_id: None | list[Literal['process', 'taskindex', 'unit', 'tag', 'config', 'run']] = None

	#: If true, a value will be set for T3/T4Document.datetime
	human_timestamp: bool = False

	#: Used if human_timestamp is true
	human_timestamp_format: str = "%Y-%m-%d %H:%M:%S.%f"


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.adapters: dict[int, AbsUnitResultAdapter] = {}


	def get_adapter_instance(self, model: UnitModel, event_hdlr: EventHandler) -> AbsUnitResultAdapter:
		config_id = build_unsafe_dict_id(model.dict())

		if (
			config_id not in self.adapters or
			self.adapters[config_id].run_id != event_hdlr.get_run_id()
		):
			unit_instance = self.context.loader.new_context_unit(
				model = model,
				context = self.context,
				run_id = event_hdlr.get_run_id(),
				sub_type = AbsUnitResultAdapter,
			)
			self.adapters[config_id] = unit_instance

		return self.adapters[config_id]

	
	def craft_doc(self,
		event_hdlr: EventHandler,
		unit: AmpelUnit,
		res: None | UBson | UnitResult,
		ts: float,
		doc_type: type[T]
	) -> T:

		d: T = {'process': event_hdlr.process_name} # type: ignore[typeddict-item]
		actact = MetaActionCode(0)
		now = datetime.now(tz=timezone.utc)

		if self.human_timestamp:
			d['datetime'] = now.strftime(self.human_timestamp_format)

		d['unit'] = unit.__class__.__name__
		d['code'] = actact

		conf = unit._get_trace_content()  # noqa: SLF001
		meta: MetaRecord = {
			'run': event_hdlr.get_run_id(),
			'ts': int(now.timestamp()),
			'duration': now.timestamp() - ts
		}

		confid = build_unsafe_dict_id(conf)
		self.context.db.add_conf_id(confid, conf)

		# Live dangerously
		if confid not in self.context.config._config['confid']:  # noqa: SLF001
			dict.__setitem__(self.context.config._config['confid'], confid, conf)  # noqa: SLF001

		d['confid'] = confid

		if self.resolve_config:
			d['config'] = conf

		if self.channel:
			d['channel'] = self.channel
			actact |= MetaActionCode.ADD_CHANNEL

		d['code'] = DocumentCode.NOT_SET
		d['meta'] = meta # note: mongodb maintains key order

		if isinstance(res, UnitResult):

			if res.code:
				d['code'] = res.code
				actact |= MetaActionCode.SET_UNIT_CODE
			else:
				actact |= MetaActionCode.SET_CODE

			if res.tag:
				if self.tag:
					d['tag'] = merge_tags(self.tag, res.tag)
				else:
					d['tag'] = res.tag
			elif self.tag:
				d['tag'] = self.tag

			if res.body:

				if res.adapter_model:
					res = self.get_adapter_instance(res.adapter_model, event_hdlr).handle(res)

				d['body'] = res.body
				actact |= MetaActionCode.ADD_BODY

		else:

			if self.tag:
				d['tag'] = self.tag

			# bson
			if isinstance(res, ubson):
				d['body'] = res
				actact |= (MetaActionCode.ADD_BODY | MetaActionCode.SET_CODE)

			else:
				actact |= MetaActionCode.SET_CODE

		meta['activity'] = [{'action': actact}]

		if self.human_id:
			ids = []
			if 'process' in self.human_id:
				ids.append(f"[{event_hdlr.process_name}]")
			if 'taskindex' in self.human_id:
				ids.append("[#{}]".format(event_hdlr.process_name.split("#")[-1]))
			if 'unit' in self.human_id:
				ids.append(f"[{unit.__class__.__name__}]")
			if 'tag' in self.human_id and d.get('tag'):
				ids.append("[%s]" % (d['tag'] if isinstance(d['tag'], int | str) \
					else " ".join(d['tag']))) # type: ignore[arg-type]
			if 'config' in self.human_id:
				ids.append(f"[{confid}]")
			if 'run' in self.human_id:
				ids.append(f"[{event_hdlr.get_run_id()}]")
			ids.append(datetime.now(tz=timezone.utc).strftime(self.human_timestamp_format))
			d['_id'] = " ".join(ids)

		return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/DataLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Iterable, Iterator
from typing import Literal

from bson.codec_options import CodecOptions

from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.utils import safe_query_dict
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.mongo.query.general import build_general_query
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.types import ChannelId, StockId, StrictIterable, Tag
from ampel.util.collections import ampel_iter

# Monitoring counters
stat_db_loads = AmpelMetricsRegistry.counter(
	"loads", "Number of documents loaded",
	subsystem="db", labelnames=("col",)
)

class DataLoader:

	def __init__(self, ctx: AmpelContext) -> None:
		self.ctx = ctx

	def load(self,
		stock_ids: StockId | Iterator[StockId] | StrictIterable[StockId],
		directives: Iterable[LoaderDirective],
		channel: None | ChannelId | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None,
		tag: None | dict[Literal['with', 'without'], Tag | dict | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag]] = None,
		auto_project: bool = True,
		codec_options: None | CodecOptions = CodecOptions(document_class=FrozenValuesDict), # noqa: B008
		logger: None | AmpelLogger = None
	) -> Iterable[AmpelBuffer]:
		"""
		:param directives: see LoaderDirective docstrings for more information.  Notes:
		- T3 unit configurations can use aliased directives. For example, the alias COMPOUND
		is translated into the directive LoaderDirective(col="t1")
		- LoaderDirective can contain "query_complement", which can be used among other thing to further
		sub-select t2 documents or select given states (such as the latest state)

		:param tag: If specified, query selection critera will apply to all directives.

		:param channel: If specified, query selection critera will apply to all directives (except t0).
		None means all channels are considered (no criterium).

		:param auto_project: each LoaderDirective is associated with a db collection with is itself associated with
		a default "content" typed dict. By default, we request a projection that projects only the fields defined
		in those TypedDict. Custom/admin fields would thus not be retrieved.
		Set this setting to False if it is not the whished behavior.
		"""

		col_set = {directive.col for directive in directives}

		# Note: the following operation will consume
		# stock_ids if it is an Iterator/Cursor
		register: dict[StockId, AmpelBuffer] = {
			stock_id: AmpelBuffer(
				id = stock_id,
				stock = None,
				t0 = [] if "t0" in col_set else None,
				t1 = [] if "t1" in col_set else None,
				t2 = [] if "t2" in col_set else None
			) for stock_id in ampel_iter(stock_ids)
		}

		for directive in directives:

			query = build_general_query(
				stock=register.keys(), channel=channel, tag=tag
			)

			#if directive.col == "stock":
			#	query['_id'] = query.pop("stock")
			#elif directive.col == "t0" and channel:
			#	query.pop('channel')

			# query 'stock' parameter primes over query complements
			if directive.query_complement:
				query = directive.query_complement | query

			if logger and logger.verbose > 1: # log query parameters
				logger.debug(
					None, extra={
						'col': directive.col,
						'query': safe_query_dict(query, dict_key=None)
					}
				)

			# Retrieve pymongo cursor
			col = self.ctx.db.get_collection(directive.col)

			if codec_options:
				col = col.database.get_collection(col.name, codec_options=codec_options)

			# Note: codec_options freezes structures in dicts with depth level > 1
			cursor = col.find(
				filter = query,
				projection = {
					k: 1 for k in directive.model.__annotations__
				} if auto_project else None
			)

			inc = stat_db_loads.labels(directive.col).inc

			if directive.col == "t1":
				count = 0
				for count, res in enumerate(cursor, 1): # noqa: B007
					register[res['stock']][directive.col].append(res) # type: ignore[union-attr]
				inc(count)

			elif directive.col == "stock":
				count = 0
				for count, res in enumerate(cursor, 1): # noqa: B007
					register[res['stock']]['stock'] = res
				inc(count)

			# Datapoints are potentially channel-less and can be associated with multiple stocks
			elif directive.col == "t0":

				count = 0
				for count, res in enumerate(cursor, 1): # noqa: B007

					# Upper limits can be attached to multiple transients
					for sid in res['stock']:

						# Some of which might not match with our query
						if sid not in register:
							continue

						# Add datapoints to snapdata (no need to recursive_freeze
						# since dict elements with level > 1 are frozen due to codec_options
						register[sid]['t0'].append(res) # type: ignore[union-attr]
				inc(count)

			elif directive.col == "t2":

				# the entire data will need to fit in memory anyway
				res = list(cursor)

				# whether to replace init config integer hash with 'resolved' config dict
				if directive.resolve_config:
					for el in res:
						dict.__setitem__(el, 'config', self.ctx.config.get_conf_id(el['config']))

				inc(len(res))

				if directive.excluding_query:

					sids = set(register.keys())
					for el in res:
						if el['stock'] in sids:
							sids.remove(el['stock'])
					if sids:
						for k in sids:
							del register[k]

				for el in res:
					register[el['stock']]['t2'].append(el) # type: ignore[union-attr]

			else:
				raise ValueError(
					f"Unrecognized LoaderDirective: {directive.dict()}"
				)

		if logger and logger.verbose:
			s = f"Unique ids: {len(register)}"
			for col_name in (col_set - set(["stock"])):
				s += f", {col_name}: "
				s += str(sum([1 for k in register for el in register[k][col_name]])) # type: ignore[misc, union-attr]
			logger.info(s)

		return register.values()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/test/dummy.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  11.02.2021
# Last Modified By:    jvs

import pathlib
import time
from collections.abc import Sequence
from typing import Any

from ampel.abstract.AbsConfigMorpher import AbsConfigMorpher
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.abstract.AbsTiedStateT2Unit import AbsTiedStateT2Unit
from ampel.abstract.AbsUnitResultAdapter import AbsUnitResultAdapter
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.core.EventHandler import EventHandler
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.ingest.IngestDirective import IngestDirective
from ampel.model.StateT2Dependency import StateT2Dependency
from ampel.struct.Resource import Resource
from ampel.struct.UnitResult import UnitResult
from ampel.types import StockId, UBson
from ampel.view.T2DocView import T2DocView


class Sleepy(AbsEventUnit):
    """
    A processor that does nothing (especially not touching the db, which is not
    mocked in subprocesses)
    """

    process_name: str = "ZZZzzzz"

    def proceed(self, event_hdlr):
        time.sleep(1)

    # override run() so as to not touch the db
    def run(self, event_hdlr):
        return self.proceed(event_hdlr)


class DummyMuxer(AbsT0Muxer):
    #: number of points to add to provided list
    points_to_insert: int = 5

    def process(
        self, dps: list[DataPoint], stock_id: None | StockId = None
    ) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
        """
        :returns: tuple[datapoints to insert, datapoints to combine]
            <datapoints to insert> will be provided to a T0 ingester
            <datapoints to combine> will potentially be provided to an underlying T1 combiner
        """

        new_dps: list[DataPoint] = [
            {"id": i, "stock": stock_id or 0}  # type: ignore[typeddict-item]
            for i in range(dps[-1]["id"] + 1, dps[-1]["id"] + 1 + self.points_to_insert)
        ]
        assert self.points_to_insert == 5
        assert len(new_dps) == self.points_to_insert
        return new_dps + dps, new_dps + dps


class DummyHistoryMuxer(AbsT0Muxer):
    def process(
        self, dps: list[DataPoint], stock_id: StockId | None = None
    ) -> tuple[list[DataPoint] | None, list[DataPoint] | None]:
        dps_db = {
            d["id"]: d
            for d in self.context.db.get_collection("t0").find({"stock": stock_id})
        }
        dps_alert = {d["id"]: d for d in dps}
        dps_to_insert = [d for d in dps if d["id"] not in dps_db]
        dps_to_combine = [
            dps_db[k] if k in dps_db else dps_alert[k] for k in set(dps_db).union(dps_alert)
        ]
        return dps_to_insert, dps_to_combine


class DummyStockT2Unit(AbsStockT2Unit):
    def process(self, stock_doc):
        return {"id": stock_doc["stock"]}


class DummyPointT2Unit(AbsPointT2Unit):
    def process(self, datapoint):
        return {"thing": datapoint["body"]["thing"]}


class DummyStateT2Unit(AbsStateT2Unit):
    foo: int = 42

    def process(self, compound, datapoints):
        return {"len": len(datapoints)}


class DummyTiedStateT2Unit(AbsTiedStateT2Unit):
    t2_dependency = [StateT2Dependency(unit="DummyStateT2Unit", link_override=None)]

    def process(
        self,
        compound: T1Document,
        datapoints: Sequence[DataPoint],
        t2views: Sequence[T2DocView],
    ) -> UBson | UnitResult:
        assert t2views, "dependencies were found"
        assert len(t2views[-1].body or []) == 1
        data = t2views[-1].get_payload() or {}
        assert isinstance(data, dict)
        return {k: v * 2 for k, v in data.items()}


class DummyCompilerOptions(CompilerOptions):
    stock: dict[str, Any] = {"id_mapper": "ZTFIdMapper", "tag": "ZTF"}
    t0: dict[str, Any] = {"tag": "ZTF"}
    t1: dict[str, Any] = {"tag": "ZTF"}
    state_t2: dict[str, Any] = {"tag": "ZTF"}
    point_t2: dict[str, Any] = {"tag": "ZTF"}
    stock_t2: dict[str, Any] = {"tag": "ZTF"}


class DummyOutputUnit(AbsEventUnit):
    value: str
    path: pathlib.Path

    def proceed(self, event_hdlr: EventHandler) -> Any:
        self.path.write_text(self.value)


class DummyInputUnit(AbsEventUnit):
    value: str
    expected_value: str

    def proceed(self, event_hdlr: EventHandler) -> Any:
        assert self.value == self.expected_value


class DummyResourceInputUnit(DummyInputUnit):
    # add an extra level of indirection via resources
    def proceed(self, event_hdlr: EventHandler) -> Any:
        assert event_hdlr.resources
        assert event_hdlr.resources[self.value].value == self.expected_value


class DummyResourceOutputUnit(AbsEventUnit):

    name: str
    value: str

    def proceed(self, event_hdlr: EventHandler) -> Any:
        event_hdlr.add_resource(Resource(name=self.name, value=self.value))


class DummyProcessorTemplate(AbsConfigMorpher):

    value: str
    expected_value: str

    def morph(self, config: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:
        return {'unit': "DummyInputUnit", 'config': self.dict(exclude={'template'})}


class DummyUnitResultAdapter(AbsUnitResultAdapter):
    def handle(self, ur: UnitResult) -> UnitResult:
        return ur


class DummyIngestUnit(AbsEventUnit):
    directives: list[IngestDirective]

    def proceed(self, event_hdlr):
        return super().proceed(event_hdlr)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/ConfigCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.07.2021
# Last Modified Date:  19.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
import os
import shutil
import subprocess
from argparse import ArgumentParser, FileType
from collections.abc import Iterable, Mapping, Sequence
from io import StringIO, UnsupportedOperation
from pathlib import Path
from time import time
from typing import Any, TextIO

import yaml

from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.config import get_user_data_config_path
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import DEBUG, INFO, AmpelLogger
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.DictSecretProvider import DictSecretProvider
from ampel.secret.PotemkinSecretProvider import PotemkinSecretProvider
from ampel.util.pretty import out_stack, prettyjson

hlp = {
	'build': 'Generate a new ampel config based on information' +
		'\n from the currently installed ampel repositories',
	'show': 'Show config / config path',
	'install': 'Build and install new config as default config for current system ' +
		'(conda envs supported).\n Option "-config" of other CLI operations becomes then optional',
	'transform': 'Transform specified config file using jq parameters',
	'validate': 'Validate all unit configurations defined a specified config file',
	'file': 'Path to an ampel config file to be installed (generation step won\'t occur)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'json': 'Show JSON encoded config',
	'pretty': 'Show pretty JSON encoded config',
	'out': 'Path to file where config will be saved',
	'path': 'Show installed config path rather than config content',
	'sign': 'Append truncated file signature (last 6 digits) to filename',
	'stop-on-errors': 'by default, config building stops and raises an exception if an error occured.\n' +
		'- 2: stop on errors\n' +
		'- 1: ignore errors in first_pass_config only (will stop on morphing/scoping/template errors)\n' +
		'- 0: ignore all errors',
	'distributions': 'Ampel packages to consider. If unspecified, gather all installed ampel distributions',
	'verbose': 'verbose',
	'ext-resource': 'path to resource config file (yaml) to be integrated into the final ampel config',
	'hide-module-not-found-errors': 'Hide ModuleNotFoundError exception stacks',
	'hide-stderr': 'Hide stderr messages arising during imports (from healpix for ex.)',
	'ignore-channels': 'ignore channel definitions',
	'ignore-processes': 'ignore process definitions',
	'no-provenance': 'Do not retrieve and save unit module dependency information\n(speeds up config building process at the detriment of traceability)'
}


class ConfigCommand(AbsCoreCommand):


	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['install', 'build', 'show', 'transform', 'validate']


	# Implement
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'config', sub_ops, hlp, description = 'Build or show ampel config.'
			)

		builder = ArgParserBuilder('config')
		ps = builder.add_parsers(sub_ops, hlp)
		ps[sub_ops.index('show')].args_not_required = True
		ps[sub_ops.index('install')].args_not_required = True

		# Required args
		builder.xargs(
			group='required', sub_ops='build', xargs = [
				{'name': 'out', 'type': str},
				{
					'name': 'install', 'action': 'store_true',
					'help': 'Installs the generated config (conda envs are supported)'
				}
			]
		)

		# Optional args
		builder.opt('secrets', 'build|install', default=None)
		builder.opt('verbose', 'build|show', action='store_true')

		builder.opt('sign', 'build', type=int, default=0)
		builder.opt('ext-resource', 'build|install')
		builder.opt('hide-module-not-found-errors', 'build|install', action='store_true')
		builder.opt('hide-stderr', 'build|install', action='store_true')
		builder.opt('ignore-channels', 'build|install', action='store_true')
		builder.opt('ignore-processes', 'build|install', action='store_true')
		builder.opt('no-provenance', 'build|install', action='store_true')
		builder.xargs(
			group='optional', sub_ops='show', xargs = [
				dict(name='json', action='store_true'),
				dict(name='path', action='store_true')
			]
		)
		builder.opt('pretty', 'show', action='store_true')
		builder.opt('stop-on-errors', 'build|install', default=2, type=int)
		builder.opt('distributions', 'build|install', nargs="+", default=["pyampel-", "ampel-"])
		builder.opt('exclude-distributions', 'build|install', nargs="+", default=[])
		builder.opt('file', 'install|validate|transform', type=FileType('r'))
		builder.opt('secrets', 'validate|transform', type=FileType('r'))
		builder.opt('out', 'transform', type=FileType('a'))
		builder.opt('filter', 'transform')
		builder.opt('validate', 'transform', action='store_true')

		# Example
		builder.example('build', '-install')
		builder.example('build', '-out ampel_conf.yaml')
		builder.example('build', '-out ampel_conf.yaml -sign -verbose')
		builder.example('build', '-out ampel_core_conf.yaml -distributions ampel-interface ampel-core')
		builder.example('show', '')
		builder.example('show', '-path')
		builder.example('show', '-json -pretty')
		builder.example('install', '')
		builder.example('install', '-stop-on-error 0 -hide-module-not-found-errors -exclude-distributions Ampel-HU-cosmo')
		builder.example('install', '-ignore-channels -ignore-processes -no-provenance')

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	@classmethod
	def _to_strict_json(cls, obj: Any) -> Any:
		""" Get JSON-compliant representation of obj """
		if isinstance(obj, Mapping):
			assert '__nonstring_keys' not in obj
			doc = {str(k): cls._to_strict_json(v) for k, v in obj.items()}
			nonstring_keys = {
				str(k): cls._to_strict_json(k) for k in obj if not isinstance(k, str)
			}
			if nonstring_keys:
				doc['__nonstring_keys'] = nonstring_keys
			return doc
		if isinstance(obj, Iterable) and not isinstance(obj, str):
			return [cls._to_strict_json(v) for v in obj]
		if isinstance(obj, int) and abs(obj) >> 53:
			# use canonical BSON representation for ints larger than the precision
			# of a double
			return {'$numberLong': str(obj)}
		return obj


	@staticmethod
	def _from_strict_json(doc):
		""" Invert to_strict_json() """
		if '$numberLong' in doc:
			return int(doc['$numberLong'])
		if '__nonstring_keys' in doc:
			nonstring_keys = doc.pop('__nonstring_keys')
			return {nonstring_keys[k]: v for k, v in doc.items()}
		return doc


	@staticmethod
	def _load_dict(source: TextIO) -> dict[str, Any]:
		if isinstance((payload := yaml.safe_load(source)), dict):
			return payload
		raise TypeError('buf does not deserialize to a dict')


	@classmethod
	def _validate(cls, config_file: TextIO, secrets: None | TextIO = None) -> None:

		from ampel.model.ChannelModel import ChannelModel
		from ampel.model.ProcessModel import ProcessModel

		ctx = AmpelContext.load(
			cls._load_dict(config_file),
			vault=AmpelVault(providers=[(
				DictSecretProvider(cls._load_dict(secrets))
				if secrets is not None
				else PotemkinSecretProvider()
			)]),
		)

		with ctx.loader.validate_unit_models():
			for channel in ctx.config.get(
				'channel', dict[str, Any], raise_exc=True
			).values():
				ChannelModel(**{k: v for k, v in channel.items() if k not in {'template'}})
			for tier in range(3):
				for process in ctx.config.get(
					f'process.t{tier}', dict[str, Any], raise_exc=True
				).values():
					ProcessModel(**process)

	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		logger = AmpelLogger.get_logger(
			console={'level': DEBUG if args.get('verbose', False) else INFO}
		)

		if sub_op == 'build':

			logger.info('Building config [use -verbose for more details]')

			# Fix ArgParserBuilder/ArgumentParser later
			if not args.get('out') and not args.get('install'):
				with out_stack():
					raise ValueError('Argument "out" or "install" required\n')

			start_time = time()
			cb = DistConfigBuilder(
				options = DisplayOptions(
					verbose = args.get('verbose', False),
					hide_stderr = args.get('hide_stderr', False),
					hide_module_not_found_errors = args.get('hide_module_not_found_errors', False)
				),
				logger = logger,
			)

			cb.load_distributions(
				prefixes=args['distributions'],
				raise_exc=args['stop_on_errors'] != 0,
				exclude=args['exclude_distributions']
			)
			cb.build_config(
				stop_on_errors = args['stop_on_errors'],
				skip_default_processes=True,
				config_validator = None,
				save = args.get('out') or get_user_data_config_path(),
				ext_resource = args.get('ext_resource'),
				sign = args.get('sign', 0),
				ignore_channels = args.get('ignore_channels', False),
				ignore_processes = args.get('ignore_processes', False),
				get_unit_env = not args.get('no_provenance', False),
			)

			dm = divmod(time() - start_time, 60)
			logger.info(
				f'Total time required: {round(dm[0])} minutes {round(dm[1])} seconds\n' 
			)

			logger.flush()

		elif sub_op == 'install':

			std_conf = Path(get_user_data_config_path())
			if not std_conf.parent.exists():
				std_conf.parent.mkdir(parents=True)
			if args['file'] and os.path.exists(args['file']):
				shutil.copy(args['file'], std_conf)
				logger.info(f'{args["file"]} successfully set as standard config ({std_conf})')
				return

			args['out'] = str(std_conf)
			self.run(args, unknown_args, sub_op = 'build')
			logger.info(f'New config built and installed ({std_conf})')

		elif sub_op == 'show':

			conf_path = get_user_data_config_path()
			if args['path']:
				print(conf_path)
				return
			if not os.path.exists(conf_path):
				logger.info(f'Config with path {conf_path} not found')
				return
			with open(conf_path) as f:
				if args['json']:
					if args['pretty']:
						print(prettyjson(yaml.safe_load(f.read())))
					else:
						print(yaml.safe_load(f.read()))
				else:
					for l in f.readlines():
						print(l, end='')

		elif sub_op == 'transform':

			try:
				with Path(args['filter']).open() as f:
					jq_args = [f.read()]
			except (FileNotFoundError, IsADirectoryError):
				jq_args = [args['filter']]

			# Use a custom transformation to losslessly round-trip from YAML to JSON,
			# in particular:
			# - wrap large ints to prevent truncation to double precision
			# - preserve non-string keys
			input_json = json.dumps(self._to_strict_json(yaml.safe_load(args['file'])))
			config = json.loads(
				subprocess.check_output(['jq', *jq_args], input=input_json.encode()),
				object_hook=self._from_strict_json,
			)
			args['file'].close()

			with StringIO() as output_yaml:
				yaml.dump(config, output_yaml, sort_keys=False)
				if args['validate']:
					output_yaml.seek(0)
					self._validate(output_yaml, args['secrets'])
				output_yaml.seek(0)
				# truncate now, in case writing back to the input file
				try:
					args['out'].seek(0)
					args['out'].truncate()
				except UnsupportedOperation:
					...
				args['out'].write(output_yaml.read())
				args['out'].close()

		elif sub_op == 'validate':
			try:
				self._validate(args['file'], args['secrets'])
			finally:
				args['file'].close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/aux/LogicOperatorFilterModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  17.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections
import operator
from collections.abc import Callable
from typing import Any, TypeAlias

from ampel.base.AmpelBaseModel import AmpelBaseModel

ops: dict[str, Callable[[str, Any], bool]] = {
	'>': operator.gt,
	'<': operator.lt,
	'>=': operator.ge,
	'<=': operator.le,
	'==': operator.eq,
	'!=': operator.ne,
	'contains': operator.contains,
	'is': operator.is_,
	'is not': operator.is_not
}

# define an alias to distinguish between type (the field name) and type (a type) in
# the annotation of `type` below
Type: TypeAlias = type

class FilterCriterion(AmpelBaseModel):

	attribute: None | str = None
	type: None | Type = None
	operator: Callable
	value: Any

	def __init__(self, **kwargs):
		if isinstance(type_kw := kwargs.get("type"), str):
			kwargs["type"] = getattr(collections.abc, type_kw)
		
		if isinstance(operator_kw := kwargs.get("operator"), str):
			if operator_kw not in ops:
				raise ValueError(f"Unknown operator: {operator_kw}")
			kwargs["operator"] = ops[operator_kw]

		super().__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/db/MongoClientRoleModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.10.2019
# Last Modified Date:  13.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel


class MongoClientRoleModel(AmpelBaseModel):
	r: str
	w: str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/secret/DictSecretProvider.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                14.08.2020
# Last Modified Date:  07.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import subprocess
from typing import Any, _GenericAlias, get_args  # type: ignore[attr-defined]

import yaml

from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.NamedSecret import NamedSecret
from ampel.secret.Secret import Secret


class DictSecretProvider(AbsSecretProvider):

	@classmethod
	def load(cls, path: str) -> 'DictSecretProvider':
		"""
		Load from a YAML file. If the file was encrypted with sops_, it will
		be decrypted with ``sops -d``.

		.. _sops: https://github.com/mozilla/sops
		"""
		with open(path) as f:
			payload = yaml.safe_load(f)

		if "sops" in payload:
			try:
				payload = yaml.safe_load(
					subprocess.check_output(['sops', '-d', path])
				)
			except Exception as exc:
				raise RuntimeError(f"Can't read sops-encrypted file {path}") from exc

		return cls(payload)


	def __init__(self, items: dict[str, Any]) -> None:
		""" """
		self.store: dict[str, Any] = dict(items)


	def tell(self, arg: Secret, ValueType: type) -> bool:
		"""
		Potentially update an initialized Secret instance with
		the actual sensitive information associable with it.
		:returns: True if the Secret was told/resolved or False
		if the provided Secret is unknown to this secret provider
		"""

		if isinstance(arg, NamedSecret) and arg.label in self.store:

			if isinstance(ValueType, _GenericAlias):
				ValueType = get_args(ValueType) # type: ignore[assignment]

			if isinstance(value := self.store[arg.label], ValueType):
				arg.set(value)
				return True

		return False



from collections.abc import Generator
from contextlib import contextmanager
from functools import partial
from typing import Any

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsIngester import AbsIngester
from ampel.content.DataPoint import DataPoint
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.model.UnitModel import UnitModel
from ampel.queue.AbsProducer import AbsProducer


class QueueIngester(AbsIngester):

    producer: UnitModel

    class QueueStockIngester(AbsDocIngester[StockDocument]):
        queue: "QueueIngester"

        def ingest(self, doc: StockDocument) -> None:
            self.queue.add_stock(doc)

    class QueueT0Ingester(AbsDocIngester[DataPoint]):
        queue: "QueueIngester"

        def ingest(self, doc: DataPoint) -> None:
            self.queue.add_t0(doc)

    class QueueT1Ingester(AbsDocIngester[T1Document]):
        queue: "QueueIngester"

        def ingest(self, doc: T1Document) -> None:
            self.queue.add_t1(doc)

    class QueueT2Ingester(AbsDocIngester[T2Document]):
        queue: "QueueIngester"

        def ingest(self, doc: T2Document) -> None:
            self.queue.add_t2(doc)

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._producer = self.context.loader.new(
            self.producer, unit_type=AbsProducer
        )

        self._stock = self.QueueStockIngester(queue=self)
        self._t0 = self.QueueT0Ingester(queue=self)
        self._t1 = self.QueueT1Ingester(queue=self)
        self._t2 = self.QueueT2Ingester(queue=self)

        self._item = AbsProducer.Item.new()
        self._messages_to_ack: list[Any] = []

    @contextmanager
    def group(self) -> Generator:
        """
        Ensure that updates issued in this context are grouped together
        """
        item = AbsProducer.Item.new()
        messages: set[Any] = set()
        prev = self._item, self._messages_to_ack
        self._item, self._messages = item, messages
        try:
            yield
        finally:
            self._item, self._messages_to_ack = prev
            self._producer.produce(
                item,
                partial(self.acknowledge_callback, iter(messages))
                if self.acknowledge_callback
                else None,
            )

    def acknowledge_on_delivery(self, message) -> None:
        self._messages_to_ack.append(message)

    def flush(self) -> None:
        self._producer.flush()

    def add_stock(self, doc: StockDocument) -> None:
        self._item.stock.append(doc)

    def add_t0(self, doc: DataPoint) -> None:
        self._item.t0.append(doc)

    def add_t1(self, doc: T1Document) -> None:
        self._item.t1.append(doc)

    def add_t2(self, doc: T2Document) -> None:
        self._item.t2.append(doc)

    @property
    def stock(self) -> AbsDocIngester[StockDocument]:
        return self._stock

    @property
    def t0(self) -> AbsDocIngester[DataPoint]:
        return self._t0

    @property
    def t1(self) -> AbsDocIngester[T1Document]:
        return self._t1

    @property
    def t2(self) -> AbsDocIngester[T2Document]:
        return self._t2




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/aux/AuxAliasableModel.py
# License           : BSD-3-Clause
# Author            : jvs
# Date              : Unspecified
# Last Modified Date: Unspecified
# Last Modified By  : jvs

from typing import Any

from pydantic import model_validator

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.model.UnitModel import UnitModel


class AuxAliasableModel(AmpelBaseModel):
	"""
	A model that can be initialized from the name of an aux unit
	"""

	@model_validator(mode="before")
	def resolve_alias(cls: type["AuxAliasableModel"], value: Any) -> dict[str, Any]:
		if isinstance(value, str):
			if value in AuxUnitRegister._defs:  # noqa: SLF001
				return AuxUnitRegister.new_unit(model=UnitModel(unit=value), sub_type=cls).model_dump()
			raise ValueError(f"{cls.__name__} '{value}' not registered")
		return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/getch.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.06.2021
# Last Modified Date:  27.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys


def yes_no(question: str) -> bool:

	try:
		while True:
			c = input(f"{question} ? [y/n]: ").lower()
			if c in ("y", "yes", "no", "n"):
				break
			sys.stdout.write('\x1b[1A')
			sys.stdout.write('\x1b[2K')
	except KeyboardInterrupt as e:
		print('\nAbording...\n\n')  # noqa: T201
		raise e

	if c in ('y', 'yes'):
		return True

	if c in ('n', 'no'):
		return False

	raise ValueError(f"Unsupported response (expected y or n): {c}")


def getch() -> bool:
	try:
		input()
	except KeyboardInterrupt:
		return True
	return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/EventCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.12.2022
# Last Modified Date:  24.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import ArgumentParser
from collections.abc import Callable, Sequence
from datetime import datetime
from typing import Any

import yaml
from bson.objectid import ObjectId

from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.util.pretty import prettyjson


class EventCommand(AbsCoreCommand):

	def __init__(self):
		self.parser = None

	@staticmethod
	def get_sub_ops() -> None | list[str]:
		return None

	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if self.parser:
			return self.parser

		parser = AmpelArgumentParser('event')
		parser.args_not_required = True
		parser.set_help_descr({
			'debug': 'Debug',
			'config': 'Path to an ampel config file',
			'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
			'jobs': 'Show only job events',
			'all': 'Show unsuccesful events as well (code != 0)',
			'resolve-jobs': 'Include resolved job schema in event documents',
			'keep-oids': 'Do not convert OIDs to datetime isoformat',
			'no-pretty': 'Do not prettify JSON',
			'yaml': 'Show YAML formatted results (potential argument value [None (default), True, False]\nwill be passed to dump parameter "default_flow_style".',
			'after': 'Match events that occured after provided date-time string [%%]',
			'before': 'Match events that occured before provided date-time string [%%]',
			'run': 'Integer run id(s) [o]',
			'run-json': 'Matching dict for run id(s) |§|',
		})

		parser.opt('config', type=str)
		parser.opt('db', default=None)
		parser.opt('debug', action='store_true')
		parser.opt('jobs', action='store_true')
		parser.opt('all', action='store_true')
		parser.opt('resolve-jobs', action='store_true')
		parser.opt('keep-oids', action='store_true')
		parser.opt('no-pretty', action='store_true')
		parser.opt('yaml', type=eval, default=-1, const='None', nargs='?')
		parser.opt('after', type=str)
		parser.opt('before', type=str)
		parser.xargs_opt(
			{'name': 'run', 'action': MaybeIntAction, 'nargs': '+'},
			{'name': 'run-json', 'action': LoadJSONAction, 'metavar': '#', 'dest': 'run'}
		)
	

		parser.hint_time_format(ref='%')
		#parser.hint_all_query_logic(ref='&')
		parser.note('multi-values are OR-matched', ref='o')
		parser.example('event -db SIM -jobs -debug')
		parser.example('event -yaml -resolve-jobs -after "2022-12-19T18:26:23+00:00"')
		parser.example('event -run-json \'{"$gt": 12}\'', ref='§')
		return parser


	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)

		ctx = self.get_context(
			args, unknown_args, ContextClass=AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag=LogFlag.MANUAL_RUN
		)

		col = ctx.db.get_collection('event', mode='r')
		if 'event' not in col.database.list_collection_names():
			logger.info(f"Event collection does not exist (db: {col.database.name})")
			return

		if args['debug']:
			logger.debug(f"Querying {col.database.name} database {col.database.client.address}")

		matchd: dict[str, Any] = {} if args['all'] else {'code': 0}

		if args['jobs']:
			matchd['jobid'] = {'$exists': True}

		if args['before']:
			self.add_time_constraint(matchd, args['before'], '$lte')

		if args['after']:
			self.add_time_constraint(matchd, args['after'], '$gte')

		if args['run']:
			if isinstance(args['run'], dict):
				matchd['run'] = args['run'] # enables -run-json '{"$gt": 1}'
			else:
				matchd['run'] = args['run'] if isinstance(args['run'], int) else {'$in': args['run']}

		sep = ','
		morphers: list[Callable[[dict[str, Any]], None]] = []

		if args['no_pretty']:
			printfunc = lambda x: print(x, end='')  # noqa: E731
		elif args['yaml'] != -1:
			printfunc = lambda x: print(yaml.dump(x, sort_keys=False, default_flow_style=args['yaml']), end='')  # noqa: E731
			sep = ''
		else:
			printfunc = lambda x: print(prettyjson(x), end='')  # noqa: E731

		if not args['keep_oids']:
			morphers.append(
				lambda x: x.__setitem__(
					'_id', x['_id'].generation_time.isoformat()
				)
			)

		if args['resolve_jobs']:
			jcol = ctx.db.get_collection('job', mode='r')
			def resolve_jobs(d):
				if jobd := next(jcol.find({'_id': d.get('jobid')}), None):
					d['job'] = jobd
					del d['jobid']

			morphers.append(resolve_jobs)

		cursor = col.find(matchd)
		if firstd := next(cursor, None):
			# Print first result without prior comma
			if morphers:
				for mph in morphers:
					mph(firstd)
				printfunc(firstd)
				for el in cursor:
					print(sep)
					for mph in morphers:
						mph(el)
					printfunc(el)
			else:
				printfunc(firstd)
				for el in cursor:
					print(sep)
					printfunc(el)

		print("")


	def add_time_constraint(self, mcrit: dict[str, Any], dt: str | datetime, op: str) -> None:
		"""
		Note: time operation is greater than / *equals*
		:param dt: either datetime object or string (datetime.fromisoformat is used)
		"""

		if isinstance(dt, datetime):
			pass
		elif isinstance(dt, str):
			dt = datetime.fromisoformat(dt)
		else:
			raise ValueError()

		if "_id" not in mcrit:
			mcrit["_id"] = {}

		mcrit["_id"][op] = ObjectId.from_datetime(dt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/AmpelLogger.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.09.2018
# Last Modified Date:  20.04.2022
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

import logging
import sys
import traceback
from os.path import basename
from sys import _getframe
from typing import TYPE_CHECKING, Any

from ampel.log.handlers.AmpelStreamHandler import AmpelStreamHandler
from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.LogFlag import LogFlag
from ampel.protocol.LoggingHandlerProtocol import (
	AggregatingLoggingHandlerProtocol,
	LoggingHandlerProtocol,
)
from ampel.types import ChannelId

if TYPE_CHECKING:
	from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler

ERROR = LogFlag.ERROR
WARNING = LogFlag.WARNING
SHOUT = LogFlag.SHOUT
INFO = LogFlag.INFO
VERBOSE = LogFlag.VERBOSE
DEBUG = LogFlag.DEBUG

if TYPE_CHECKING:
	from ampel.core.AmpelContext import AmpelContext

class AmpelLogger:

	loggers: dict[int | str, 'AmpelLogger'] = {}
	_counter: int = 0
	verbose: int = 0


	@classmethod
	def get_logger(cls, name: None | int | str = None, force_refresh: bool = False, **kwargs) -> 'AmpelLogger':
		"""
		Creates or returns an instance of :obj:`AmpelLogger <ampel.log.AmpelLogger>`
		that is registered in static dict 'loggers' using the provided name as key.
		If a logger with the given name already exists, the existing logger instance is returned.
		If name is None, unique (int) name will be generated
		:param ``**kwargs``: passed to constructor

		Typical use:\n
		.. sourcecode:: python\n
			logger = AmpelLogger.get_logger()
		"""

		if not name:
			cls._counter += 1
			name = cls._counter

		if name not in AmpelLogger.loggers or force_refresh:
			AmpelLogger.loggers[name] = AmpelLogger(name=name, **kwargs)

		return AmpelLogger.loggers[name]


	@staticmethod
	def from_profile(context: 'AmpelContext', profile: str, run_id: None | int = None, **kwargs) -> 'AmpelLogger':

		handlers = context.config.get(f'logging.{profile}', dict, raise_exc=True)
		logger = AmpelLogger.get_logger(console=False, **kwargs)

		if "db" in handlers:
			# avoid circular import
			from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler

			if run_id is None:
				raise ValueError("Parameter 'run_id' is required when log_profile requires db logging handler")

			logger.addHandler(
				DBLoggingHandler(context.db, run_id, **handlers['db'])
			)

		if "console" in handlers:
			logger.addHandler(
				AmpelStreamHandler(**handlers['console'])
			)

		return logger


	@staticmethod
	def get_console_level(context: 'AmpelContext', profile: str) -> None | int:

		handlers = context.config.get(f'logging.{profile}', dict, raise_exc=True)

		if "console" in handlers:
			if 'level' in handlers['console']:
				return handlers['console']['level']
			return LogFlag.INFO.__int__()

		return None


	@classmethod
	def has_verbose_console(cls, context: 'AmpelContext', profile: str) -> bool:

		if lvl := cls.get_console_level(context, profile):
			return lvl < INFO
		return False


	def __init__(self,
		name: int | str = 0,
		base_flag: None | LogFlag = None,
		handlers: None | list[LoggingHandlerProtocol | AggregatingLoggingHandlerProtocol] = None,
		channel: None | ChannelId | list[ChannelId] = None,
		# See AmpelStreamHandler annotations for more details
		console: None | bool | dict[str, Any] = True
	) -> None:

		self.name = name
		self.base_flag = base_flag.__int__() if base_flag else 0
		self.handlers = handlers or []
		self.channel = channel
		self.level = 0
		self.fname = _getframe().f_code.co_filename

		if console:
			self.addHandler(
				AmpelStreamHandler() if console is True else AmpelStreamHandler(**console)
			)
		else:
			self.provenance = False

		self._auto_level()


	def _auto_level(self):

		self.level = min([h.level for h in self.handlers]) if self.handlers else 0
		if self.level < INFO:
			self.verbose = 2 if self.level < VERBOSE else 1
		elif self.verbose != 0:
			self.verbose = 0


	def addHandler(self, handler: LoggingHandlerProtocol) -> None:

		self.level = min(handler.level, self.level)

		if isinstance(handler, AmpelStreamHandler) and handler.provenance:
			self.provenance = True

		if self.level < INFO:
			self.verbose = 2 if self.level < VERBOSE else 1

		self.handlers.append(handler)


	def removeHandler(self, handler: LoggingHandlerProtocol) -> None:
		self.handlers.remove(handler)
		self._auto_level()


	def get_db_logging_handler(self) -> 'None | DBLoggingHandler':
		# avoid circular import
		from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler
		for el in self.handlers:
			if isinstance(el, DBLoggingHandler):
				return el
		return None


	def break_aggregation(self) -> None:
		for el in self.handlers:
			if isinstance(el, AggregatingLoggingHandlerProtocol | AmpelStreamHandler):
				el.break_aggregation()


	def error(self, msg: str | dict[str, Any], *args,
		exc_info: None | Exception = None,
		extra: None | dict[str, Any] = None,
	):
		self.log(ERROR, msg, *args, exc_info=exc_info, extra=extra)


	def warn(self, msg: str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	):
		if self.level <= WARNING:
			self.log(WARNING, msg, *args, extra=extra)


	def info(self, msg: None | str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	) -> None:
		if self.level <= INFO:
			self.log(INFO, msg, *args, extra=extra)


	def debug(self, msg: None | str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	):
		if self.level <= DEBUG:
			self.log(DEBUG, msg, *args, extra=extra)


	def handle(self, record: LightLogRecord | logging.LogRecord) -> None:
		for h in self.handlers:
			if record.levelno >= h.level:
				h.handle(record)


	def flush(self) -> None:
		for h in self.handlers:
			h.flush()


	def log(self,
		lvl: int, msg: None | str | dict[str, Any], *args,
		exc_info: None | bool | Exception = None,
		extra: None | dict[str, Any] = None,
	):

		if args and isinstance(msg, str):
			msg = msg % args

		record = LightLogRecord(name=self.name, levelno=lvl | self.base_flag, msg=msg)

		if lvl > WARNING or self.provenance:
			frame = _getframe(1) # logger.log(...) was called directly
			if frame.f_code.co_filename == self.fname:
				frame = _getframe(2) # logger.info(...), logger.debug(...) was used
			record.__dict__['filename'] = basename(frame.f_code.co_filename)
			record.__dict__['lineno'] = frame.f_lineno

		if extra:
			extra = dict(extra)
			if (stock := extra.pop("stock", None)):
				record.stock = stock
			if (channel := (extra.pop("channel", None) or self.channel)):
				record.channel = channel
			if (unit := (extra.pop("unit", None))):
				record.unit = unit
			record.extra = extra

		if exc_info:

			if exc_info == 1:
				exc_info = sys.exc_info() # type: ignore[assignment]
				lines = traceback.format_exception(*sys.exc_info())
			elif isinstance(exc_info, tuple):
				lines = traceback.format_exception(*sys.exc_info())
			elif isinstance(exc_info, Exception):
				lines = traceback.format_exception(
					type(exc_info), exc_info, exc_info.__traceback__
				)
			else:
				lines = []

			erec = AmpelLogger.fork_rec(record, "\n")
			for h in self.handlers:
				h.handle(erec)

			for el in lines:
				for l in el.split('\n'):
					if not l:
						continue
					erec = AmpelLogger.fork_rec(record, l)
					for h in self.handlers:
						h.handle(erec)

			if record.msg:
				rec2 = AmpelLogger.fork_rec(record, "-" * len(record.msg))
				for h in self.handlers:
					h.handle(record)
					h.handle(rec2)

			return

		for h in self.handlers:
			if record.levelno >= h.level:
				h.handle(record)


	@staticmethod
	def fork_rec(orig: LightLogRecord, msg: str) -> LightLogRecord:
		new_rec = LightLogRecord(name=0, msg=None, levelno=0)
		for k, v in orig.__dict__.items():
			new_rec.__dict__[k] = v
		new_rec.msg = msg
		return new_rec



from collections.abc import Callable, Generator, Iterator
from contextlib import contextmanager
from typing import Any

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.core.ContextUnit import ContextUnit
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.types import Traceless


class AbsIngester(AmpelABC, ContextUnit, abstract=True):
    error_callback: Traceless[None | Callable[[], None]] = None
    acknowledge_callback: Traceless[None | Callable[[Iterator[Any]], None]] = None

    raise_exc: bool = False

    run_id: Traceless[int]
    logger: Traceless[LoggerProtocol]

    @contextmanager
    @abstractmethod
    def group(self) -> Generator:
        """
        Ensure that documents ingested in this context are grouped together
        """
        ...

    @abstractmethod
    def acknowledge_on_delivery(self, message: Any) -> None:
        """
        Add a message to be passed to acknowledge_callback when the current
        batch of documents has been delivered
        """
        ...

    @abstractmethod
    def flush(self) -> None:
        """
        Wait for all documents to be stored
        """
        ...

    @property
    @abstractmethod
    def stock(self) -> AbsDocIngester[StockDocument]: ...

    @property
    @abstractmethod
    def t0(self) -> AbsDocIngester[DataPoint]: ...

    @property
    @abstractmethod
    def t1(self) -> AbsDocIngester[T1Document]: ...

    @property
    @abstractmethod
    def t2(self) -> AbsDocIngester[T2Document]: ...



import gc
import signal
from collections.abc import Generator, Sequence
from time import time
from typing import Any, Literal, TypedDict, overload

from bson import ObjectId
from mongomock.filtering import filter_applies

from ampel.abstract.AbsIngester import AbsIngester
from ampel.abstract.AbsWorker import stat_time, stop_on_signal
from ampel.content.DataPoint import DataPoint
from ampel.content.MetaRecord import MetaRecord
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.core.EventHandler import EventHandler
from ampel.enum.DocumentCode import DocumentCode
from ampel.log import AmpelLogger, LogFlag
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.queue.AbsConsumer import AbsConsumer
from ampel.t2.T2Worker import T2Worker
from ampel.types import (
	DataPointId,
	OneOrMany,
	StockId,
	T2Link,
	Tag,
	UBson,
)


class QueueItem(TypedDict):
	stock: Sequence[StockDocument]
	t0: Sequence[DataPoint]
	t1: Sequence[T1Document]
	t2: Sequence[T2Document]



class T2QueueWorker(T2Worker):

	consumer: UnitModel
	ingester: UnitModel = UnitModel(unit="MongoIngester")
	
	# Must run dependent t2s, as no other worker will see them
	run_dependent_t2s: Literal[True] = True
	pre_check: Literal[False] = False

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)

		self._current_item: None | QueueItem = None
	
	def proceed(self, event_hdlr: EventHandler) -> int:
		""" :returns: number of t2 docs processed """

		event_hdlr.set_tier(self.tier)
		run_id = event_hdlr.get_run_id()

		logger = AmpelLogger.from_profile(
			self.context, self.log_profile, run_id,
			base_flag = getattr(LogFlag, f'T{self.tier}') | LogFlag.CORE | self.base_log_flag
		)

		if self.send_beacon:
			self.col_beacon.update_one(
				{'_id': self.beacon_id},
				{'$set': {'timestamp': int(time())}}
			)

		stock_updr = MongoStockUpdater(
			ampel_db = self.context.db, tier = self.tier, run_id = run_id,
			process_name = self.process_name, logger = logger,
			raise_exc = self.raise_exc, extra_tag = self.jtag
		)

		consumer = self.context.loader.new(self.consumer, unit_type=AbsConsumer)

		# Loop variables
		doc_counter = 0
		garbage_collect = self.garbage_collect
		doc_limit = self.doc_limit

		with stop_on_signal(
			[signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGHUP],
			logger
		) as stop_token:
			ingester = self.context.loader.new_context_unit(
				self.ingester,
				context = self.context,
				run_id = run_id,
				error_callback = stop_token.set,
				acknowledge_callback = consumer.acknowledge,
				logger = logger,
				sub_type = AbsIngester,
			)

			try:
				self._current_run_id = run_id
				# Process docs until next() returns None (breaks condition below)
				while not stop_token.is_set():

					# get t1/t2 document (code is usually NEW or NEW_PRIO), excluding
					# docs with retry times in the future
					with stat_time.labels(self.tier, "consume", None).time():
						item: None | QueueItem = consumer.consume()

					# No match
					if item is None:
						if not stop_token.is_set():
							logger.log(LogFlag.SHOUT, "No more docs to process")
						break

					self._current_item = item

					for doc in item["t2"]:
						with stat_time.labels(self.tier, "process_doc", doc["unit"]).time():
							self.process_doc(doc, stock_updr, logger)
					doc_counter += 1

					with ingester.group():
						for stock in item["stock"]:
							ingester.stock.ingest(stock)
						for dp in item["t0"]:
							ingester.t0.ingest(dp)
						for t1 in item["t1"]:
							ingester.t1.ingest(t1)
						for t2 in item["t2"]:
							ingester.t2.ingest(t2)
						ingester.acknowledge_on_delivery(item)

					# Check possibly defined doc_limit
					if doc_limit and doc_counter >= doc_limit:
						break

					if garbage_collect:
						gc.collect()
			finally:
				stock_updr.flush()
				ingester.flush()
				event_hdlr.add_extra(docs=doc_counter)

				logger.flush()
				self._instances.clear()
				self._adapters.clear()
				self._current_run_id = None

		return doc_counter

	def update_doc(self,
		doc: T2Document,
		meta: MetaRecord,
		logger: AmpelLogger, *,
		body: UBson = None,
		tag: None | OneOrMany[Tag] = None,
		code: int = 0
	) -> None:
		# update doc in place; changes will be propagated by the ingester as a block
		doc["code"] = code
		doc["meta"] = [*doc["meta"], meta]
		if body:
			doc["body"] = [*doc["body"], body]
		if tag:
			doc["tag"] = list(
				set([tag] if isinstance(tag, Tag) else tag)
				.union(doc.get("tag", []))
			)

	def load_stock(self, stock: StockId) -> None | StockDocument:
		"""Load stock document from current message"""
		return (
			next((d for d in self._current_item["stock"] if d["stock"] == stock), None)
			if self._current_item is not None else None
		) or super().load_stock(stock)

	@overload
	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: DataPointId) -> None | DataPoint: ...

	@overload
	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: Sequence[DataPointId]) -> list[DataPoint]: ...

	def load_t0(self, stock: StockId | Sequence[StockId], t1_dps_ids: DataPointId | Sequence[DataPointId]) -> None | DataPoint | list[DataPoint]:
		"""Load datapoints from database"""
		if isinstance(t1_dps_ids, DataPointId):
			return (
				next((d for d in self._current_item["t0"] if d["stock"] == stock and d["id"] == t1_dps_ids), None)
				if self._current_item is not None else None
			) or super().load_t0(stock, t1_dps_ids)
		targets = set(t1_dps_ids)
		datapoints = (
			{dp["id"]: dp for dp in self._current_item["t0"] if dp["stock"] == stock and dp["id"] in targets}
			if self._current_item is not None else {}
		)
		# fall back to database if some datapoints are missing
		if missing := targets.difference(datapoints.keys()):
			return list(datapoints.values()) + super().load_t0(stock, list(missing))
		return list(datapoints.values())

	def load_t1(self, stock: StockId | Sequence[StockId], link: T2Link) -> None | T1Document:
		"""Load T1 document from database"""
		return (
			next((d for d in self._current_item["t1"] if d["stock"] == stock and d["link"] == link), None)
			if self._current_item is not None else None
		) or super().load_t1(stock, link)
	
	def load_t2(self, query: dict[str, Any], for_update: bool=False) -> Generator[T2Document]:
		"""
		Load T2 documents from database
								  |  for_update  |
		condition                 | true | false |
		--------------------------+------+-------+
		not in db                 | item | item  |
		in db, code OK            | -    | db    |
		in db, code in code_match | item | item  |
		in db, other code         | item | item  |
		"""

		if self._current_item is not None:
			if for_update:
				# want to run dependent t2, but already resolved
				if self.col.count_documents({"code": DocumentCode.OK} | query):
					return
				for doc in self._current_item["t2"]:
					if doc["code"] != DocumentCode.OK and filter_applies(query, doc):
						# prevent this doc from being returned by a call with for_update=False
						if "_id" not in doc:
							doc["_id"] = ObjectId()  # type: ignore[typeddict-unknown-key]
						yield doc
			else:
				count = 0
				for count, doc in enumerate(super().load_t2(query, for_update), 1):  # noqa: B007
					yield doc
				if count == 0:
					# return t2 docs that are not in the database
					for doc in self._current_item["t2"]:
						if filter_applies(query, doc):
							yield doc
		else:
			yield from super().load_t2(query, for_update)
